# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
import os.path

from pyaedt import Desktop
from pyaedt import Hfss
from pyaedt import Icepak
from pyaedt import Maxwell3d
from pyaedt import Q3d
from pyaedt import settings
from pyaedt.application.design_solutions import solutions_types
from pyaedt.generic.design_types import get_pyaedt_app
from pyaedt.generic.filesystem import search_files
from pyaedt.generic.general_methods import generate_unique_name
from pyaedt.workflows.misc import get_aedt_version
from pyaedt.workflows.misc import get_arguments
from pyaedt.workflows.misc import get_port
from pyaedt.workflows.misc import get_process_id
from pyaedt.workflows.misc import is_student

settings.use_grpc_api = True
settings.use_multi_desktop = True
non_graphical = True
extension_arguments = {"password": "", "application": "HFSS", "solution": "Modal", "file_path": ""}
extension_description = "Convert File from 22R2"
port = get_port()
version = get_aedt_version()
aedt_process_id = get_process_id()
is_student = is_student()


def frontend():  # pragma: no cover

    import tkinter
    from tkinter import filedialog
    from tkinter import ttk

    master = tkinter.Tk()

    master.geometry("750x250")

    master.title("Convert File from 22R2")

    # Configure style for ttk buttons
    style = ttk.Style()
    style.configure("Toolbutton.TButton", padding=6, font=("Helvetica", 8))

    var2 = tkinter.StringVar()
    label2 = tkinter.Label(master, textvariable=var2)
    var2.set("Browse file or folder:")
    label2.grid(row=0, column=0, pady=10)
    text = tkinter.Text(master, width=40, height=1)
    text.grid(row=0, column=1, pady=10, padx=5)

    def edit_sols(self):
        sol["values"] = tuple(solutions_types[appl.get()].keys())
        sol.current(0)

    var = tkinter.StringVar()
    label = tkinter.Label(master, textvariable=var)
    var.set("Password (Encrypted 3D Component Only):")
    label.grid(row=1, column=0, pady=10)
    pwd = tkinter.Entry(master, width=20, show="*")
    pwd.insert(tkinter.END, "")
    pwd.grid(row=1, column=1, pady=10, padx=5)

    var = tkinter.StringVar()
    label = tkinter.Label(master, textvariable=var)
    var.set("Application (3D Component Only):")
    label.grid(row=2, column=0, pady=10)
    appl = ttk.Combobox(master, width=40, validatecommand=edit_sols)  # Set the width of the combobox
    appl["values"] = ("HFSS", "Q3D Extractor", "Maxwell 3D", "Icepak")
    appl.current(0)
    appl.bind("<<ComboboxSelected>>", edit_sols)
    appl.grid(row=2, column=1, pady=10, padx=5)

    var = tkinter.StringVar()
    label = tkinter.Label(master, textvariable=var)
    var.set("Solution (3D Component Only):")
    label.grid(row=3, column=0, pady=10)
    sol = ttk.Combobox(master, width=40)  # Set the width of the combobox
    sol["values"] = ttk.Combobox(master, width=40)  # Set the width of the combobox
    sol["values"] = tuple(solutions_types["HFSS"].keys())
    sol.current(0)
    sol.grid(row=3, column=1, pady=10, padx=5)

    def browseFiles():
        filename = filedialog.askopenfilename(
            initialdir="/",
            title="Select a Electronics File",
            filetypes=(("AEDT", ".aedt *.a3dcomp"), ("all files", "*.*")),
        )
        text.insert(tkinter.END, filename)

    b1 = tkinter.Button(master, text="...", width=10, command=browseFiles)
    b1.grid(row=0, column=2, pady=10)

    def callback():
        applications = {"HFSS": 0, "Icepak": 1, "Maxwell 3D": 2, "Q3D Extractor": 3}
        master.password_ui = pwd.get()
        master.application_ui = applications[appl.get()]
        master.solution_ui = sol.get()
        master.file_path_ui = text.get("1.0", tkinter.END).strip()
        master.destroy()

    b3 = tkinter.Button(master, text="Ok", width=40, command=callback)
    b3.grid(row=5, column=1, pady=10, padx=10)

    tkinter.mainloop()

    password_ui = getattr(master, "password_ui", extension_arguments["password"])
    application_ui = getattr(master, "application_ui", extension_arguments["application"])
    solution_ui = getattr(master, "solution_ui", extension_arguments["solution"])
    file_path_ui = getattr(master, "file_path_ui", extension_arguments["file_path"])

    output_dict = {
        "password": password_ui,
        "application": application_ui,
        "solution": solution_ui,
        "file_path": file_path_ui,
    }
    return output_dict


def check_missing(input_object, output_object, file_path):
    if output_object.design_type not in [
        "HFSS",
        "Icepak",
        "Q3d",
        "2D Extractor",
        "Maxwell 3D",
        "Maxwell 2D",
        "Mechanical",
    ]:
        return
    object_list = input_object.modeler.object_names[::]
    new_object_list = output_object.modeler.object_names[::]
    un_classified_objects = output_object.modeler.unclassified_names[::]
    unclassified = [i for i in object_list if i not in new_object_list and i in un_classified_objects]
    disappeared = [i for i in object_list if i not in new_object_list and i not in un_classified_objects]
    list_of_suppressed = [["Design", "Object", "Operation"]]
    for obj_name in unclassified:
        if obj_name in output_object.modeler.object_names:
            continue
        hist = output_object.modeler[obj_name].history()
        for el_name, el in list(hist.children.items())[::-1]:
            if "Suppress Command" in el.props:
                el.props["Suppress Command"] = True
                list_of_suppressed.append([output_object.design_name, obj_name, el_name])
            if obj_name in output_object.modeler.object_names:
                break
    for obj_name in disappeared:
        input_object.export_3d_model(
            file_name=obj_name,
            file_format=".x_t",
            file_path=input_object.working_directory,
            assignment_to_export=[obj_name],
        )
        output_object.modeler.import_3d_cad(os.path.join(input_object.working_directory, obj_name + ".x_t"))
        list_of_suppressed.append([output_object.design_name, obj_name, "History"])
    from pyaedt.generic.general_methods import read_csv
    from pyaedt.generic.general_methods import write_csv

    if file_path.split(".")[1] == "a3dcomp":
        output_csv = os.path.join(file_path[:-8], "Import_Errors.csv")[::-1].replace("\\", "_", 1)[::-1]
    else:
        output_csv = os.path.join(file_path[:-5], "Import_Errors.csv")[::-1].replace("\\", "_", 1)[::-1]
    if os.path.exists(output_csv):
        data_read = read_csv(output_csv)
        list_of_suppressed = data_read + list_of_suppressed[1:]
    write_csv(output_csv, list_of_suppressed)
    print(f"Errors saved in {output_csv}")
    return output_csv, True


def convert_3d_component(
    extension_args,
    output_desktop,
    input_desktop,
):

    file_path = extension_args["file_path"]
    password = extension_args["password"]
    solution = extension_args["solution"]
    application = extension_args["application"]

    output_path = file_path[:-8] + f"_{version}.a3dcomp"

    if os.path.exists(output_path):
        output_path = file_path[:-8] + generate_unique_name(f"_version", n=2) + ".a3dcomp"
    app = Hfss
    if application == 1:
        app = Icepak
    elif application == 2:
        app = Maxwell3d
    elif application == 3:
        app = Q3d
    app1 = app(aedt_process_id=input_desktop.aedt_process_id, solution_type=solution)
    cmp = app1.modeler.insert_3d_component(file_path, password=password)
    app_comp = cmp.edit_definition(password=password)
    design_name = app_comp.design_name
    app_comp.oproject.CopyDesign(design_name)
    project_name2 = generate_unique_name("Proj_convert")
    output_app = app(aedt_process_id=output_desktop.aedt_process_id, solution_type=solution, project=project_name2)

    output_app.oproject.Paste()
    output_app = get_pyaedt_app(desktop=output_desktop, project_name=project_name2, design_name=design_name)
    check_missing(app_comp, output_app, file_path)
    output_app.modeler.create_3dcomponent(
        output_path,
        is_encrypted=True if password else False,
        allow_edit=True if password else False,
        edit_password=password,
        password_type="InternalPassword" if password else "UserSuppliedPassword",
        hide_contents=False,
    )
    try:
        output_desktop.DeleteProject(project_name2)
        print(f"Project successfully deleted")
    except Exception:
        print(f"Error project was not closed")
    print(f"3D Component {output_path} has been created.")


def convert_aedt(
    extension_args,
    output_desktop,
    input_desktop,
):

    file_path = extension_args["file_path"]

    file_path = str(file_path)
    a3d_component_path = str(file_path)
    output_path = a3d_component_path[:-5] + f"_{version}.aedt"
    if os.path.exists(output_path):
        output_path = a3d_component_path[:-5] + generate_unique_name(f"_{version}", n=2) + ".aedt"

    input_desktop.load_project(file_path)
    project_name = os.path.splitext(os.path.split(file_path)[-1])[0]
    oproject2 = output_desktop.odesktop.NewProject(output_path)
    project_name2 = os.path.splitext(os.path.split(output_path)[-1])[0]

    for design in input_desktop.design_list():

        app1 = get_pyaedt_app(desktop=input_desktop, project_name=project_name, design_name=design)
        app1.oproject.CopyDesign(app1.design_name)
        oproject2.Paste()
        output_app = get_pyaedt_app(desktop=output_desktop, project_name=project_name2, design_name=design)
        check_missing(app1, output_app, file_path)
        output_app.save_project()
    input_desktop.odesktop.CloseProject(os.path.splitext(os.path.split(file_path)[-1])[0])


def convert(args):
    logger = logging.getLogger("Global")
    if os.path.isdir(args["file_path"]):
        files_path = search_files(args["file_path"], "*.a3dcomp")
        files_path += search_files(args["file_path"], "*.aedt")
    else:
        files_path = [args["file_path"]]
    output_desktop = Desktop(
        new_desktop=True,
        version=version,
        port=port,
        aedt_process_id=aedt_process_id,
        student_version=is_student,
    )
    input_desktop = Desktop(new_desktop=True, version=222, non_graphical=non_graphical)
    for file in files_path:
        try:
            args["file_path"] = file
            if args["file_path"].endswith("a3dcomp"):
                convert_3d_component(args, output_desktop, input_desktop)
            else:
                convert_aedt(args, output_desktop, input_desktop)
        except Exception:
            logger.error(f"Failed to convert {file}")
    input_desktop.release_desktop()
    output_desktop.release_desktop(False, False)


if __name__ == "__main__":
    args = get_arguments(extension_arguments, extension_description)
    # Open UI
    if not args["is_batch"]:  # pragma: no cover
        output = frontend()
        if output:
            for output_name, output_value in output.items():
                if output_name in extension_arguments:
                    args[output_name] = output_value
    convert(args)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from copy import copy
import itertools
import os
import re
import subprocess

from pyaedt.generic.general_methods import is_ironpython
from pyaedt.misc.misc import installed_versions

if not is_ironpython:
    import matplotlib.pyplot as plt
    import numpy as np
    import skrf as rf

from pyaedt.generic.general_methods import pyaedt_function_handler

REAL_IMAG = "RI"
MAG_ANGLE = "MA"
DB_ANGLE = "DB"

keys = {REAL_IMAG: ("real", "imag"), MAG_ANGLE: ("mag", "deg"), DB_ANGLE: ("db20", "deg")}


def _parse_ports_name(file):
    """Parse and interpret the option line in the touchstone file.

    Parameters
    ----------
    file : str
        Path of the touchstone file.

    Returns
    -------
    List of str
        Names of the ports in the touchstone file.

    """
    portnames = []
    line = file.readline()
    while not line.startswith("! Port"):
        line = file.readline()
    while line.startswith("! Port"):
        portnames.append(line.split(" = ")[1].strip())
        line = file.readline()
    return portnames


class TouchstoneData(rf.Network):
    """Contains data information from Touchstone Read call."""

    def __init__(self, solution_data=None, touchstone_file=None):
        if solution_data is not None:
            self.solution_data = solution_data
            freq_points = solution_data.primary_sweep_values
            ports = []
            for i in solution_data.expressions:
                m = re.search(r"S\(\s*(\S+)\s*,\s*(\S+)\s*\)", i)
                if m:
                    ports.append(m.group(1))
                    ports.append(m.group(2))

            ports = list(set(ports))
            ports.sort()
            port_order = {ports.index(p): p for p in ports}
            sdata_3d = np.zeros([len(freq_points), len(ports), len(ports)], dtype=complex)
            for expression in solution_data.expressions:
                m = re.search(r"S\(\s*(\S+)\s*,\s*(\S+)\s*\)", expression)
                if m:
                    p_a = m.group(1)
                    p_b = m.group(2)
                p_a_number = ports.index(p_a)
                p_b_number = ports.index(p_b)
                sdata_real = solution_data.data_real(expression, True)
                sdata_img = solution_data.data_imag(expression, True)
                sdata_2d = np.array(sdata_real, dtype=complex) + 1j * np.array(sdata_img, dtype=complex)
                sdata_3d[:, p_a_number, p_b_number] = sdata_2d
                sdata_3d[:, p_b_number, p_a_number] = sdata_2d

            var = {}
            for name, value in solution_data.active_variation.items():
                var[name] = value
            params = {"variant": var, "port_order": port_order}
            rf.Network.__init__(self, s=sdata_3d, frequency=freq_points, z0=50.0 + 0.0j, params=params)
            rf.stylely()
            self.port_names = ports

        elif os.path.exists(touchstone_file):
            rf.Network.__init__(self, touchstone_file)
        self.log_x = True

    @pyaedt_function_handler()
    def get_insertion_loss_index(self, threshold=-3):
        """Get all insertion losses. The first frequency point is used to determine whether two
        ports are shorted.

        Parameters
        ----------
        threshold : float, int, optional
            Threshold to determine shorted ports in dB.

        Returns
        -------
         list
            List of index couples representing Insertion Losses of excitations.

        """
        temp_list = []
        s_db = self.s_db[0:2, :, :]
        for i in self.port_tuples:
            if i[0] != i[1]:
                loss = s_db[0, i[0], i[1]]
                if loss > threshold:
                    temp_list.append(i)
                elif loss < -90:
                    loss = s_db[1, i[0], i[1]]
                    if loss > threshold:
                        temp_list.append(i)
        return temp_list

    def plot_insertion_losses(self, threshold=-3, plot=True):
        """Plot all insertion losses. The first frequency point is used to determine whether two
        ports are shorted.

        Parameters
        ----------
        threshold : float, int, optional
            Threshold to determine shorted ports in dB.
        plot : bool, optional
            Whether to plot. The default is ``True``.

        Returns
        -------
        list
            List of tuples representing insertion loss excitations.
        """
        temp_list = self.get_insertion_loss_index(threshold=threshold)
        if plot:  # pragma: no cover
            for i in temp_list:
                self.plot_s_db(*i, logx=self.log_x)
            plt.show()
        return temp_list

    def plot(self, index_couples=None, show=True):
        """Plot a list of curves.

        Parameters
        ----------
        index_couples : list, optional
            List of indexes couple to plot. Default is ``None`` to plot all ``port_tuples``.
        show : bool
            Whether to plot. Default is ``True``.

        Returns
        -------
        :class:`matplotlib.plt`
        """

        if not index_couples:
            index_couples = self.port_tuples[:]

        for i in index_couples:
            self.plot_s_db(*i, logx=self.log_x)
        if show:
            plt.show()
        return plt

    def plot_return_losses(self):  # pragma: no cover
        """Plot all return losses.

        Returns
        -------
        bool
        """
        for i in np.arange(self.number_of_ports):
            self.plot_s_db(i, i, logx=self.log_x)
        plt.show()
        return True

    def get_mixed_mode_touchstone_data(self, num_of_diff_ports=None, port_ordering="1234"):
        """Transform network from single ended parameters to generalized mixed mode parameters.

        Example 1, an N-port single-ended network with port order 1234 is converted to mixed-mode
        parameters.

             A                                  B
             +------------+                     +-----------+
           0-|s0========s2|-2                 0-|d0=======d1|-1
           1-|s1========s3|-3                 2-|d2=======d3|-3
            ...          ...     =se2gmm=>     ...         ...
        2N-4-|s2N-4==s2N-2|-2N-2           2N-4-|cN-4===cN-3|-2N-3
        2N-3-|s2N-3==s2N-1|-2N-1           2N-2-|cN-2===cN-1|-2N-1
             +------------+                     +-----------+

        Example 2, an N-port single-ended network with port order 1324 is converted to mixed-mode
        parameters.

             A                                  B
             +------------+                     +-----------+
           0-|s0========s2|-1                 0-|d0=======d1|-1
           2-|s1========s3|-3                 2-|d2=======d3|-3
            ...          ...     =se2gmm=>     ...         ...
        2N-4-|s2N-4==s2N-2|-2N-3           2N-4-|cN-4===cN-3|-2N-3
        2N-2-|s2N-3==s2N-1|-2N-1           2N-2-|cN-2===cN-1|-2N-1
             +------------+                     +-----------+

        Parameters
        ----------
        num_of_diff_ports : int, optional
            The number of differential ports.
        port_ordering : str, optional
            The current port ordering. Options are ``"1234"``, ``"1324"``. The default
            is ``1234``

        Returns
        -------
        class:`pyaedt.generic.touchstone_parser.TouchstoneData`

        """
        ts_diff = copy(self)
        port_count = len(ts_diff.port_names)

        if num_of_diff_ports is None:
            num_of_diff_ports = port_count // 4 * 2

        if port_ordering == "1234":
            pass
        elif port_ordering == "1324":
            temp_port_order = np.arange(port_count)
            for i in np.arange(port_count):
                if i % 4 == 1:
                    temp_port_order[i] = temp_port_order[i] + 1
                elif i % 4 == 2:
                    temp_port_order[i] = temp_port_order[i] - 1

            port_order = np.arange(port_count)
            new_port_order = np.arange(port_count)
            new_port_order[: len(temp_port_order)] = temp_port_order

            ts_diff.renumber(port_order, new_port_order)
        else:
            return False

        ts_diff.se2gmm(num_of_diff_ports)

        new_port_names = ["D{}".format(i) for i in np.arange(num_of_diff_ports)]
        new_port_names.extend(["C{}".format(i) for i in np.arange(num_of_diff_ports)])
        ts_diff.port_names[: len(new_port_names)] = new_port_names
        return ts_diff

    @pyaedt_function_handler()
    def get_return_loss_index(self, excitation_name_prefix=""):
        """Get the list of all the Returnloss from a list of exctitations.

        If no excitation is provided it will provide a full list of return losses.

        Example: excitation_names ["1","2"] is_touchstone_expression=False output ["S(1,1)", S(2,2)]
        Example: excitation_names ["S(1,1)","S(1,2)", S(2,2)] is_touchstone_expression=True output ["S(1,1)", S(2,2)]

        Parameters
        ----------

        excitation_name_prefix :
             (Default value = '')

        Returns
        -------
        list
            List of index couples representing return losses of excitations.

        """
        values = []
        if excitation_name_prefix:
            excitation_names = [i for i in self.port_names if excitation_name_prefix.lower() in i.lower()]
        else:
            excitation_names = [i for i in self.port_names]
        for i in excitation_names:
            values.append([self.port_names.index(i), self.port_names.index(i)])
        return values

    @pyaedt_function_handler()
    def get_insertion_loss_index_from_prefix(self, tx_prefix, rx_prefix):
        """Get the list of all the Insertion Losses from prefix.

        Parameters
        ----------
        tx_prefix : str
            Prefix for TX (eg. "DIE").
        rx_prefix : str
            Prefix for RX (eg. "BGA").

        Returns
        -------
        list
            List of index couples representing Insertion Losses of excitations.

        """
        trlist = [i for i in self.port_names if tx_prefix in i]
        receiver_list = [i for i in self.port_names if rx_prefix in i]
        values = []
        if len(trlist) != len(receiver_list):
            print("TX and RX should be same length lists")
            return False
        for i, j in zip(trlist, receiver_list):
            values.append([self.port_names.index(i), self.port_names.index(j)])
        return values

    @pyaedt_function_handler()
    def get_next_xtalk_index(self, tx_prefix=""):
        """Get the list of all the Near End XTalk a list of excitation. Optionally prefix can
        be used to retrieve driver names.
        Example: excitation_names ["1", "2", "3"] output ["S(1,2)", "S(1,3)", "S(2,3)"].

        Parameters
        ----------
        tx_prefix :
            prefix for TX (eg. "DIE") (Default value = "")

        Returns
        -------
        list
            List of index couples representing Near End XTalks.

        """
        if tx_prefix:
            trlist = [i for i in self.port_names if tx_prefix in i]
        else:
            trlist = self.port_names
        values = []
        for i in trlist:
            k = trlist.index(i) + 1
            while k < len(trlist):
                values.append([self.port_names.index(i), self.port_names.index(trlist[k])])
                k += 1
        return values

    @pyaedt_function_handler()
    def get_fext_xtalk_index_from_prefix(self, tx_prefix, rx_prefix, skip_same_index_couples=True):
        """Get the list of all the Far End XTalk from a list of exctitations and a prefix that will
        be used to retrieve driver and receivers names.
        If skip_same_index_couples is true, the tx and rx with same index
        position will be considered insertion losses and excluded from the list.

        Parameters
        ----------
        tx_prefix : str
            prefix for TX (eg. "DIE")
        rx_prefix : str
            prefix for RX (eg. "BGA")
        skip_same_index_couples : bool
            Boolean ignore TX and RX couple with same index. The default value is ``True``.

        Returns
        -------
        list
            List of index couples representing Far End XTalks.

        """
        trlist = [i for i in self.port_names if tx_prefix in i]
        reclist = [i for i in self.port_names if rx_prefix in i]
        values = []
        for i in trlist:
            for k in reclist:
                if not skip_same_index_couples or reclist.index(k) != trlist.index(i):
                    values.append([self.port_names.index(i), self.port_names.index(k)])
        return values

    def plot_next_xtalk_losses(self, tx_prefix=""):
        """Plot all next crosstalk curves.

        Parameters
        ----------
        Returns
        -------
        bool
        """
        index = self.get_next_xtalk_index(tx_prefix=tx_prefix)

        for ind in index:
            self.plot_s_db(ind[0], ind[1], logx=self.log_x)
        plt.show()
        return True

    def plot_fext_xtalk_losses(self, tx_prefix, rx_prefix, skip_same_index_couples=True):
        """Plot all fext crosstalk curves.

        Parameters
        ----------
        tx_prefix : str
            prefix for TX (eg. "DIE")
        rx_prefix : str
            prefix for RX (eg. "BGA")
        skip_same_index_couples : bool
            Boolean ignore TX and RX couple with same index. The default value is ``True``.

        Returns
        -------
        bool
        """
        index = self.get_fext_xtalk_index_from_prefix(
            tx_prefix=tx_prefix, rx_prefix=rx_prefix, skip_same_index_couples=skip_same_index_couples
        )
        for ind in index:
            self.plot_s_db(ind[0], ind[1], logx=self.log_x)
        plt.show()
        return True

    @pyaedt_function_handler()
    def get_worst_curve(self, freq_min=None, freq_max=None, worst_is_higher=True, curve_list=None, plot=True):
        """Analyze a solution data object with multiple curves and find the worst curve.
        Take the mean of the magnitude over the frequency range.

        Parameters
        ----------
        freq_min : float, optional
            Minimum frequency to analyze in GHz (None to 0). Default value is ``None``.
        freq_max : float, optional
            Maximum frequency to analyze in GHz (None to max freq). Default value is ``None``.
        worst_is_higher : bool
            Worst curve is the one with higher mean value. Default value is ``True``.
        curve_list : list
            List of [m,n] index of curves on which to search. None to search on all curves. Default value is ``None``.
        plot : bool, optional
            Whether to plot or not the chart.

        Returns
        -------
        tuple
            Worst element, dictionary of ordered expression.

        """

        return_loss_freq = [float(i.center) for i in list(self.frequency)]
        if not freq_min:
            lower_id = 0
        else:
            lower_id = next(x[0] for x in enumerate(return_loss_freq) if x[1] >= freq_min * 1e9)
        if not freq_max:
            higher_id = len(return_loss_freq) - 1
        else:
            if freq_max * 1e9 >= return_loss_freq[-1]:
                higher_id = len(return_loss_freq) - 1
            else:
                higher_id = next(x[0] for x in enumerate(return_loss_freq) if x[1] >= freq_max * 1e9)

        dict_means = {}
        if not curve_list:
            curve_list = list(itertools.combinations(range(0, len(self.port_names)), 2))
            for i in range(0, len(self.port_names)):
                curve_list.append([i, i])
        for el in curve_list:
            data1 = np.absolute(self.s[lower_id:higher_id, el[0], el[1]])
            mean1 = sum(data1) / len(data1)
            dict_means[tuple(el)] = mean1
        dict_means = dict(sorted(dict_means.items(), key=lambda item: item[1], reverse=worst_is_higher))
        worst_el = next(iter(dict_means))
        if plot:  # pragma: no cover
            self.plot_s_db(*worst_el, logx=self.log_x)
            plt.show()
        return worst_el, dict_means


@pyaedt_function_handler(file_path="input_file")
def read_touchstone(input_file):
    """Load the contents of a Touchstone file into an NPort.

    Parameters
    ----------
    input_file : str
        The path of the touchstone file.

    Returns
    -------
    class:`pyaedt.generic.touchstone_parser.TouchstoneData`
        NPort holding data contained in the touchstone file.

    """
    data = TouchstoneData(touchstone_file=input_file)
    return data


@pyaedt_function_handler(folder="input_dir")
def check_touchstone_files(input_dir="", passivity=True, causality=True):
    """Check passivity and causality for all Touchstone files included in the folder.

    Parameters
    ----------
    input_dir : str
        Folder path. The default is ``""``.
    passivity : bool, optional
        Whether the passivity check is enabled, The default is ``True``.
    causality : bool, optional
        Whether the causality check is enabled. The default is ``True``.

    Returns
    ----------
    dict
        Dictionary with the SNP file name as the key and a list if the passivity and/or causality checks are enabled.
        The first element in the list is a str with ``"passivity"`` or ``"causality"`` as a value. The second element
        is a Boolean that is set to ``True`` when the criteria passed or ``False`` otherwise. The last element
        is a string with the log information.

    """
    out = {}
    sNpFiles = find_touchstone_files(input_dir)
    if not sNpFiles:
        return out
    aedt_install_folder = list(installed_versions().values())[0]
    for snpf in sNpFiles:
        out[snpf] = []
        if os.name == "nt":
            genequiv_path = os.path.join(aedt_install_folder, "genequiv.exe")
        else:
            genequiv_path = os.path.join(aedt_install_folder, "genequiv")
        cmd = [genequiv_path]
        if passivity:
            cmd.append("-checkpassivity")
        if causality:
            cmd.append("-checkcausality")

        cmd.append(sNpFiles[snpf])
        my_env = os.environ.copy()
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=my_env)  # nosec
        output = p.communicate()
        output_str = str(output[0])
        output_lst = output_str.split("\\r\\n")

        if len(output_lst) == 1:
            output_lst = output_str.splitlines()
        for line in output_lst:
            if "Input data" in line and passivity:
                msg_log = line[17:]
                is_passive = True
                if "non-passive" in msg_log:
                    is_passive = False
                out[snpf].append(["passivity", is_passive, msg_log])
            if "Maximum causality" in line and causality:
                msg_log = line[17:]
                is_causal = True
                try:
                    causality_check = float(msg_log.split("Maximum causality error: ")[-1].split("for entry")[0])
                    if not causality_check == 0.0:
                        is_causal = False
                except Exception:
                    is_causal = False
                    raise Exception("Failed evaluating causality value.")
                out[snpf].append(["causality", is_causal, msg_log])
            if "Causality check is inconclusive" in line and causality:
                is_causal = False
                out[snpf].append(["causality", is_causal, line[17:]])
    return out


@pyaedt_function_handler()
def find_touchstone_files(input_dir):
    """Get all Touchstone files in a directory.

    Parameters
    ----------
    input_dir : str
        Folder path. The default is ``""``.

    Returns
    ----------
    dict
        Dictionary with the SNP file names as the key and the absolute path as the value.
    """
    out = {}
    if not os.path.exists(input_dir):  # pragma: no cover
        return out
    pat_snp = re.compile(r"\.s\d+p$", re.IGNORECASE)
    sNpFiles = {f: os.path.join(input_dir, f) for f in os.listdir(input_dir) if re.search(pat_snp, f)}
    pat_ts = re.compile("\.ts$")
    for f in os.listdir(input_dir):
        if re.search(pat_ts, f):
            sNpFiles[f] = os.path.abspath(os.path.join(input_dir, f))
    return sNpFiles



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ctypes import POINTER
from ctypes import byref
from ctypes import c_char_p
from ctypes import c_int
from ctypes import create_string_buffer

import pyaedt


class MultipleBandsTable:
    """Manipulates access to the entries of multiple bands table.

    This class allows you to enter, edit, or remove entries in the multiple bands table.
    The table includes the lower and upper frequencies of the bands.
    To access the multiple bands table, use the ``multiple_bands_table`` attribute of the ``FilterSolutions`` class.
    A valid multiple bands table must include both lower and upper frequencies for each band.
    """

    def __init__(self):
        self._dll = pyaedt.filtersolutions_core._dll_interface()._dll
        self._dll_interface = pyaedt.filtersolutions_core._dll_interface()
        self._define_multiple_bands_dll_functions()

    def _define_multiple_bands_dll_functions(self):
        """Define C++ API DLL functions."""
        self._dll.getMultipleBandsTableRowCount.argtype = POINTER(c_int)
        self._dll.getMultipleBandsTableRowCount.restype = c_int

        self._dll.getMultipleBandsTableRow.argtypes = [c_int, c_char_p, c_char_p, c_int]
        self._dll.getMultipleBandsTableRow.restype = c_int

        self._dll.updateMultipleBandsTableRow.argtypes = [c_int, c_char_p, c_char_p]
        self._dll.updateMultipleBandsTableRow.restype = c_int

        self._dll.appendMultipleBandsTableRow.argtypes = [c_char_p, c_char_p]
        self._dll.appendMultipleBandsTableRow.restype = c_int

        self._dll.insertMultipleBandsTableRow.argtypes = [c_int, c_char_p, c_char_p]
        self._dll.insertMultipleBandsTableRow.restype = c_int

        self._dll.removeMultipleBandsTableRow.argtype = c_int
        self._dll.removeMultipleBandsTableRow.restype = c_int

    @property
    def row_count(self) -> int:
        """Total number of rows present in the multiple bands table.
        The default is ``2``.

        Returns
        -------
        int
            Current number of rows in the multiple bands table.
        """
        table_row_count = c_int()
        status = self._dll.getMultipleBandsTableRowCount(byref(table_row_count))
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)
        return int(table_row_count.value)

    def row(self, row_index):
        """Get the lower and upper frequency values for a row in the multiple bands table.

        Parameters
        ----------
        row_index: int
            Index of the row from to get the frequency values from.
            Valid values range from ``0`` to ``6``, inclusive.

        Returns
        -------
            tuple
                The tuple contains three strings.The first is the lower frequency value,
                and the second is the upper frequency value.

        """
        lower_value_buffer = create_string_buffer(100)
        upper_value_buffer = create_string_buffer(100)
        status = self._dll.getMultipleBandsTableRow(row_index, lower_value_buffer, upper_value_buffer, 100)
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)
        lower_value_string = lower_value_buffer.value.decode("utf-8")
        upper_value_string = upper_value_buffer.value.decode("utf-8")
        return lower_value_string, upper_value_string

    def update_row(self, row_index, lower_frequency="", upper_frequency=""):
        """Update lower and upper frequency values for a row in the multiple bands table.

        Parameters
        ----------
        row_index: int
            Index of the row to update. Valid values range from ``0`` to ``6``, inclusive.
        lower_frequency: str, optional
            New lower frequency value to set for the row.
            If this value is not provided, the row's lower frequency remains unchanged.
        upper_frequency: str, optional
            New upper frequency value to set for the row.
            If this value is not provided, the row's upper frequency remains unchanged.
        """
        lower_bytes_value = bytes(lower_frequency, "ascii")
        upper_bytes_value = bytes(upper_frequency, "ascii")
        status = self._dll.updateMultipleBandsTableRow(row_index, lower_bytes_value, upper_bytes_value)
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)

    def append_row(self, lower_frequency, upper_frequency):
        """Append a new row with specified lower and upper frequency values to the end of the multiple bands table.

        Parameters
        ----------
        lower_frequency: str
            Lower frequency value for the new row.
        upper_frequency: str
            Upper frequency value for the new row.
        """
        lower_bytes_value = bytes(lower_frequency, "ascii")
        upper_bytes_value = bytes(upper_frequency, "ascii")
        status = self._dll.appendMultipleBandsTableRow(lower_bytes_value, upper_bytes_value)
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)

    def insert_row(self, row_index, lower_frequency, upper_frequency):
        """Insert lower and upper frequencies in a given row.

        Parameters
        ----------
        row_index: int
            Index of the row. Valid values range from ``0`` to ``6``, inclusive.
        lower_frequency: str
            Lower frequency value to insert.
        upper_frequency: str
            Upper frequency value to insert.
        """
        lower_bytes_value = bytes(lower_frequency, "ascii")
        upper_bytes_value = bytes(upper_frequency, "ascii")
        status = self._dll.insertMultipleBandsTableRow(row_index, lower_bytes_value, upper_bytes_value)
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)

    def remove_row(self, row_index):
        """Remove a row specified by its index from the multiple bands table.

        Parameters
        ----------
        row_index: int
            The index of the row to be removed from the multiple bands table.
            Valid values range from ``0``to ``6``, inclusive.
        """
        status = self._dll.removeMultipleBandsTableRow(row_index)
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import csv
import os

import pyaedt
from pyaedt import Icepak
import pyaedt.workflows
from pyaedt.workflows.misc import get_aedt_version
from pyaedt.workflows.misc import get_arguments
from pyaedt.workflows.misc import get_port
from pyaedt.workflows.misc import get_process_id
from pyaedt.workflows.misc import is_student

port = get_port()
version = get_aedt_version()
aedt_process_id = get_process_id()
is_student = is_student()

# Extension batch arguments
extension_arguments = {"file_path": ""}
extension_description = "Power map from file"


def frontend():  # pragma: no cover

    import tkinter
    from tkinter import filedialog
    from tkinter import ttk

    import PIL.Image
    import PIL.ImageTk

    app = pyaedt.Desktop(
        new_desktop=False,
        version=version,
        port=port,
        aedt_process_id=aedt_process_id,
        student_version=is_student,
    )

    active_project = app.active_project()
    active_design = app.active_design()

    project_name = active_project.GetName()

    if active_design.GetDesignType() == "Icepak":
        design_name = active_design.GetName()
    else:  # pragma: no cover
        app.logger.error("Icepak project is needed.")
        app.release_desktop(False, False)
        raise Exception("Icepak 3D Layout project is needed.")

    ipk = Icepak(project_name, design_name)

    # Create UI
    master = tkinter.Tk()

    master.geometry("600x150")

    master.title("Create power maps from csv file")

    # Load the logo for the main window
    icon_path = os.path.join(pyaedt.workflows.__path__[0], "images", "large", "logo.png")
    im = PIL.Image.open(icon_path)
    photo = PIL.ImageTk.PhotoImage(im)

    # Set the icon for the main window
    master.iconphoto(True, photo)

    # Configure style for ttk buttons
    style = ttk.Style()
    style.configure("Toolbutton.TButton", padding=6, font=("Helvetica", 8))
    var2 = tkinter.StringVar()
    label2 = tkinter.Label(master, textvariable=var2)
    var2.set("Browse file:")
    label2.grid(row=1, column=0, pady=10)
    text = tkinter.Text(master, width=50, height=1)
    text.grid(row=1, column=1, pady=10, padx=5)

    def browseFiles():
        filename = filedialog.askopenfilename(
            initialdir="/",
            title="Select csv file",
            filetypes=(("power map", "*.csv*"), ("all files", "*.*")),
        )
        text.insert(tkinter.END, filename)

    b1 = tkinter.Button(master, text="...", width=10, command=browseFiles)
    b1.grid(row=3, column=0)
    b1.grid(row=1, column=2, pady=10)

    def callback():
        master.file_path_ui = text.get("1.0", tkinter.END).strip()
        master.destroy()

    b = tkinter.Button(master, text="Create", width=40, command=callback)
    b.grid(row=2, column=1, pady=10)

    tkinter.mainloop()

    file_path_ui = getattr(master, "file_path_ui", extension_arguments["file_path"])

    if not file_path_ui or not os.path.isfile(file_path_ui):
        app.logger.error("File does not exist.")

    ipk.release_desktop(False, False)

    output_dict = {"file_path": file_path_ui}

    return output_dict


def main(extension_args):

    csv_file = extension_args["file_path"]
    app = pyaedt.Desktop(
        new_desktop=False,
        version=version,
        port=port,
        aedt_process_id=aedt_process_id,
        student_version=is_student,
    )
    active_project = app.active_project()
    active_design = app.active_design()
    project_name = active_project.GetName()

    if active_design.GetDesignType() in ["Icepak"]:
        design_name = active_design.GetName()
    else:  # pragma: no cover
        app.logger.debug("Icepak project is needed.")
        app.release_desktop(False, False)
        raise Exception("Icepak 3D Layout project is needed.")

    ipk = Icepak(project_name, design_name)

    # read csv file

    create_powermaps_from_csv(ipk, csv_file)

    if not extension_args["is_test"]:  # pragma: no cover
        ipk.release_desktop(False, False)
    return True


def create_powermaps_from_csv(ipk, csv_file):
    """
    Creates powermap from an Icepak classic CSV file.

    Parameters
    ----------

    csv_file : str
        The file path to the CSV file to be processed.

    """

    geometric_info, source_value_info, source_unit_info = extract_info(csv_file)
    create_powermaps_from_info(ipk, geometric_info, source_value_info, source_unit_info)


def create_powermaps_from_info(ipk, geometric_info, source_value_info, source_unit_info):
    """
    Creates power maps from geometric and source information.

     Parameters
     ----------

    ipk:
     geometric_info : list
         A list of dictionaries, each containing:
             - "name": The name of the geometric object.
             - "vertices": A list of vertex coordinates.
    source_value_info: dict
         A dictionary mapping geometric object to its power value.
    source_unit_info: dict
         A dictionary mapping geometric object to its power unit.

    """

    for info in geometric_info:
        name = info["name"]
        points = []
        for vertex in info["vertices"]:
            if not vertex == "":
                x = vertex.split()[0] + "m"
                y = vertex.split()[1] + "m"
                z = vertex.split()[2] + "m"
                points.append([x, y, z])
        # add first point at the end of list
        points.append(points[0])
        ipk.logger.info("creating 2d object " + name)
        sanitized_name = name.replace(".", "_")
        polygon = ipk.modeler.create_polyline(points, name=sanitized_name)
        ipk.logger.info("created polygon " + polygon.name)
        ipk.modeler.cover_lines(polygon)
        power = source_value_info[name] + source_unit_info[name]
        ipk.logger.info("Assigning power value " + power)
        ipk.assign_source(polygon.name, "Total Power", power)
        ipk.logger.info("Assigned power value " + power)


def extract_info(csv_file):
    """
    Extracts source and geometric information from an Icepak classic CSV file.

    Parameters
    ----------

    csv_file (str): The file path to the CSV file to be processed.

    Returns
    -------
    geometric_info : list
        A list of dictionaries, each containing:
            - "name": The name of the geometric object.
            - "vertices": A list of vertex coordinates.
    source_value_info: dict
        A dictionary mapping geometric object to its power value.
    source_unit_info: dict
        A dictionary mapping geometric object to its power unit.

    """
    # Initialize lists to store the extracted information
    source_value_info = {}
    source_unit_info = {}
    geometric_info = []

    with open(csv_file, "r") as file:
        reader = csv.reader(file)

        # Skip the first three header lines
        for _ in range(3):
            next(reader)

        # Read the source information lines until an empty line
        for line in reader:
            if not line or line[0] == "":
                break
            source_value_info[line[0]] = line[1]
            source_unit_info[line[0]] = line[2]

        # Skip the next three lines
        for _ in range(3):
            next(reader)

        # Read the geometric information
        for line in reader:
            if line and line[0]:
                geometric_info.append({"name": line[0], "vertices": line[10:]})

        return geometric_info, source_value_info, source_unit_info


if __name__ == "__main__":  # pragma: no cover
    args = get_arguments(extension_arguments, extension_description)

    # Open UI
    if not args["is_batch"]:  # pragma: no cover
        output = frontend()
        if output:
            for output_name, output_value in output.items():
                if output_name in extension_arguments:
                    args[output_name] = output_value

    main(args)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ctypes import POINTER
from ctypes import byref
from ctypes import c_bool
from ctypes import c_char_p
from ctypes import c_int

import pyaedt


class LumpedNodesandLeads:
    """Defines attributes of the lumped element node capacitors and lead inductors.

    This class lets you to construct all the necessary node capacitors and
    lead inductors attributes of the lumped elements for the ``LumpedDesign`` class.
    """

    def __init__(self):
        self._dll = pyaedt.filtersolutions_core._dll_interface()._dll
        self._dll_interface = pyaedt.filtersolutions_core._dll_interface()
        self._define_nodes_and_leads_dll_functions()

    def _define_nodes_and_leads_dll_functions(self):
        """Define C++ API DLL functions."""
        self._dll.setLumpedCNodeCapacitor.argtype = c_char_p
        self._dll.setLumpedCNodeCapacitor.restype = c_int
        self._dll.getLumpedCNodeCapacitor.argtypes = [c_char_p, c_int]
        self._dll.getLumpedCNodeCapacitor.restype = c_int

        self._dll.setLumpedCLeadInductor.argtype = c_char_p
        self._dll.setLumpedCLeadInductor.restype = c_int
        self._dll.getLumpedCLeadInductor.argtypes = [c_char_p, c_int]
        self._dll.getLumpedCLeadInductor.restype = c_int

        self._dll.setLumpedLNodeCapacitor.argtype = c_char_p
        self._dll.setLumpedLNodeCapacitor.restype = c_int
        self._dll.getLumpedLNodeCapacitor.argtypes = [c_char_p, c_int]
        self._dll.getLumpedLNodeCapacitor.restype = c_int

        self._dll.setLumpedLLeadInductor.argtype = c_char_p
        self._dll.setLumpedLLeadInductor.restype = c_int
        self._dll.getLumpedLLeadInductor.argtypes = [c_char_p, c_int]
        self._dll.getLumpedLLeadInductor.restype = c_int

        self._dll.setLumpedRNodeCapacitor.argtype = c_char_p
        self._dll.setLumpedRNodeCapacitor.restype = c_int
        self._dll.getLumpedRNodeCapacitor.argtypes = [c_char_p, c_int]
        self._dll.getLumpedRNodeCapacitor.restype = c_int

        self._dll.setLumpedRLeadInductor.argtype = c_char_p
        self._dll.setLumpedRLeadInductor.restype = c_int
        self._dll.getLumpedRLeadInductor.argtypes = [c_char_p, c_int]
        self._dll.getLumpedRLeadInductor.restype = c_int

        self._dll.setLumpedCNodeLedComensate.argtype = c_bool
        self._dll.setLumpedCNodeLedComensate.restype = c_int
        self._dll.getLumpedCNodeLedComensate.argtype = POINTER(c_bool)
        self._dll.getLumpedCNodeLedComensate.restype = c_int

        self._dll.setLumpedLNodeLedComensate.argtype = c_bool
        self._dll.setLumpedLNodeLedComensate.restype = c_int
        self._dll.getLumpedLNodeLedComensate.argtype = POINTER(c_bool)
        self._dll.getLumpedLNodeLedComensate.restype = c_int

    @property
    def c_node_capacitor(self) -> str:
        """Shunt capacitance assigned to each capacitor node.
        The default is ``0``.

        Returns
        -------
        str
        """
        c_node_capacitor = self._dll_interface.get_string(self._dll.getLumpedCNodeCapacitor)
        return c_node_capacitor

    @c_node_capacitor.setter
    def c_node_capacitor(self, c_node_capacitor):
        self._dll_interface.set_string(self._dll.setLumpedCNodeCapacitor, c_node_capacitor)

    @property
    def c_lead_inductor(self) -> str:
        """Series inductance assigned to each capacitor lead.
        The default is ``0``.

        Returns
        -------
        str
        """
        c_lead_inductor = self._dll_interface.get_string(self._dll.getLumpedCLeadInductor)
        return c_lead_inductor

    @c_lead_inductor.setter
    def c_lead_inductor(self, c_lead_inductor):
        self._dll_interface.set_string(self._dll.setLumpedCLeadInductor, c_lead_inductor)

    @property
    def l_node_capacitor(self) -> str:
        """Shunt capacitance assigned to each inductor node.
        The default is` ``0``.

        Returns
        -------
        str
        """
        l_node_capacitor = self._dll_interface.get_string(self._dll.getLumpedLNodeCapacitor)
        return l_node_capacitor

    @l_node_capacitor.setter
    def l_node_capacitor(self, l_node_capacitor):
        self._dll_interface.set_string(self._dll.setLumpedLNodeCapacitor, l_node_capacitor)

    @property
    def l_lead_inductor(self) -> str:
        """Series inductance assigned to each inductor lead.
        The default is ``0``.

        Returns
        -------
        str
        """
        l_lead_inductor = self._dll_interface.get_string(self._dll.getLumpedLLeadInductor)
        return l_lead_inductor

    @l_lead_inductor.setter
    def l_lead_inductor(self, l_lead_inductor):
        self._dll_interface.set_string(self._dll.setLumpedLLeadInductor, l_lead_inductor)

    @property
    def r_node_capacitor(self) -> str:
        """Shunt capacitance assigned to each resistor node.
        The default is ``0``.

        Returns
        -------
        str
        """
        r_node_capacitor = self._dll_interface.get_string(self._dll.getLumpedRNodeCapacitor)
        return r_node_capacitor

    @r_node_capacitor.setter
    def r_node_capacitor(self, r_node_capacitor):
        self._dll_interface.set_string(self._dll.setLumpedRNodeCapacitor, r_node_capacitor)

    @property
    def r_lead_inductor(self) -> str:
        """Series inductance assigned to each resistor lead.
        The default is ``0``.

        Returns
        -------
        str
        """
        r_lead_inductor = self._dll_interface.get_string(self._dll.getLumpedRLeadInductor)
        return r_lead_inductor

    @r_lead_inductor.setter
    def r_lead_inductor(self, r_lead_inductor):
        self._dll_interface.set_string(self._dll.setLumpedRLeadInductor, r_lead_inductor)

    @property
    def c_node_compensate(self) -> bool:
        """Flag indicating if the adjusting capacitor values to compensate for node capacitance is enabled.

        Returns
        -------
        bool
        """
        c_node_compensate = c_bool()
        status = self._dll.getLumpedCNodeLedComensate(byref(c_node_compensate))
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)
        return bool(c_node_compensate.value)

    @c_node_compensate.setter
    def c_node_compensate(self, c_node_compensate: bool):
        status = self._dll.setLumpedCNodeLedComensate(c_node_compensate)
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)

    @property
    def l_node_compensate(self) -> bool:
        """Flag indicating if the adjusting inductor values to compensate for lead inductance is enabled.

        Returns
        -------
        bool
        """
        l_node_compensate = c_bool()
        status = self._dll.getLumpedLNodeLedComensate(byref(l_node_compensate))
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)
        return bool(l_node_compensate.value)

    @l_node_compensate.setter
    def l_node_compensate(self, l_node_compensate: bool):
        status = self._dll.setLumpedLNodeLedComensate(l_node_compensate)
        pyaedt.filtersolutions_core._dll_interface().raise_error(status)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ctypes import c_char_p
from ctypes import c_int

import pyaedt


class GraphSetup:
    """Defines the frequency and time limits of the exported responses."""

    def __init__(self):
        self._dll = pyaedt.filtersolutions_core._dll_interface()._dll
        self._dll_interface = pyaedt.filtersolutions_core._dll_interface()
        self._define_graph_dll_functions()

    def _define_graph_dll_functions(self):
        """Define C++ API DLL functions."""
        self._dll.setPlotMinimumFrequency.argtype = c_char_p
        self._dll.setPlotMinimumFrequency.restype = c_int
        self._dll.getPlotMinimumFrequency.argtypes = [c_char_p, c_int]
        self._dll.getPlotMinimumFrequency.restype = c_int

        self._dll.setPlotMaximumFrequency.argtype = c_char_p
        self._dll.setPlotMaximumFrequency.restype = c_int
        self._dll.getPlotMaximumFrequency.argtypes = [c_char_p, c_int]
        self._dll.getPlotMaximumFrequency.restype = c_int

        self._dll.setPlotMinimumTime.argtype = c_char_p
        self._dll.setPlotMinimumTime.restype = c_int
        self._dll.getPlotMinimumTime.argtypes = [c_char_p, c_int]
        self._dll.getPlotMinimumTime.restype = c_int

        self._dll.setPlotMaximumTime.argtype = c_char_p
        self._dll.setPlotMaximumTime.restype = c_int
        self._dll.getPlotMaximumTime.argtypes = [c_char_p, c_int]
        self._dll.getPlotMaximumTime.restype = c_int

    @property
    def minimum_frequency(self) -> str:
        """Minimum frequency value for exporting frequency and S parameter responses. The default is ``200 MHz``.

        Returns
        -------
        str
        """
        min_freq_string = self._dll_interface.get_string(self._dll.getPlotMinimumFrequency)
        return min_freq_string

    @minimum_frequency.setter
    def minimum_frequency(self, min_freq_string):
        self._dll_interface.set_string(self._dll.setPlotMinimumFrequency, min_freq_string)

    @property
    def maximum_frequency(self) -> str:
        """Maximum frequency value for exporting frequency and S parameters responses. The default is ``5 GHz``.

        Returns
        -------
        str
        """
        max_freq_string = self._dll_interface.get_string(self._dll.getPlotMaximumFrequency)
        return max_freq_string

    @maximum_frequency.setter
    def maximum_frequency(self, max_freq_string):
        self._dll_interface.set_string(self._dll.setPlotMaximumFrequency, max_freq_string)

    @property
    def minimum_time(self) -> str:
        """Minimum time value for exporting time response. The default is ``0 s``.

        Returns
        -------
        str
        """
        min_time_string = self._dll_interface.get_string(self._dll.getPlotMinimumTime)
        return min_time_string

    @minimum_time.setter
    def minimum_time(self, min_time_string):
        self._dll_interface.set_string(self._dll.setPlotMinimumTime, min_time_string)

    @property
    def maximum_time(self) -> str:
        """Maximum time value for exporting time response. The default is ``10 ns``.

        Returns
        -------
        str
        """
        max_time_string = self._dll_interface.get_string(self._dll.getPlotMaximumTime)
        return max_time_string

    @maximum_time.setter
    def maximum_time(self, max_time_string):
        self._dll_interface.set_string(self._dll.setPlotMaximumTime, max_time_string)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pyaedt
from pyaedt.filtersolutions_core.attributes import Attributes
from pyaedt.filtersolutions_core.attributes import FilterImplementation
from pyaedt.filtersolutions_core.graph_setup import GraphSetup
from pyaedt.filtersolutions_core.ideal_response import IdealResponse
from pyaedt.filtersolutions_core.lumped_nodes_and_leads import LumpedNodesandLeads
from pyaedt.filtersolutions_core.lumped_parasitics import LumpedParasitics
from pyaedt.filtersolutions_core.lumped_termination_impedance import LumpedTerminationImpedance
from pyaedt.filtersolutions_core.lumped_termination_impedance import TerminationType
from pyaedt.filtersolutions_core.lumped_topology import LumpedTopology
from pyaedt.filtersolutions_core.multiple_bands_table import MultipleBandsTable
from pyaedt.filtersolutions_core.transmission_zeros import TableFormat
from pyaedt.filtersolutions_core.transmission_zeros import TransmissionZeros


class FilterSolutions:
    """Provides the ``FilterSolutions`` application interface.

    This class has access to ideal filter attributes and calculated output parameters.

    Parameters
    ----------
    version: str, optional
        Version of AEDT in ``xxxx.x`` format. The default is ``None``.
    implementation_type: FilterImplementation, optional
        Technology used to implement the filter. The default is ``LUMPED``.
        The ``FilterImplementation`` enum provides the list of implementations.


    Examples
    --------
    Create a ``FilterSolutions`` instance with a band-pass elliptic ideal filter.

    >>> import pyaedt
    >>> from pyaedt.filtersolutions_core.attributes import FilterImplementation

    >>> design = pyaedt.FilterSolutions(version="2025 R1", projectname= "fs1",
    >>> implementation_type= FilterImplementation.LUMPED,
    >>> )
    """

    def __init__(self, version=None, implementation_type=None):
        self.version = version
        self.implementation_type = implementation_type
        pyaedt.filtersolutions_core._dll_interface(version)

        if implementation_type == FilterImplementation.LUMPED or implementation_type is None:
            self._init_lumped_design()
        else:
            raise RuntimeError("The " + str(implementation_type) + " is not supported in this release.")

    def _init_lumped_design(self):
        """Initialize the ``FilterSolutions`` object to support a lumped filter design."""

        self.attributes = Attributes()
        self.ideal_response = IdealResponse()
        self.graph_setup = GraphSetup()
        self.topology = LumpedTopology()
        self.parasitics = LumpedParasitics()
        self.leads_and_nodes = LumpedNodesandLeads()
        self.source_impedance_table = LumpedTerminationImpedance(TerminationType.SOURCE)
        self.load_impedance_table = LumpedTerminationImpedance(TerminationType.LOAD)
        self.multiple_bands_table = MultipleBandsTable()
        self.transmission_zeros_ratio = TransmissionZeros(TableFormat.RATIO)
        self.transmission_zeros_bandwidth = TransmissionZeros(TableFormat.BANDWIDTH)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import OrderedDict
import json
import math
import os
import shutil
import sys
import time

from pyaedt.aedt_logger import pyaedt_logger as logger
from pyaedt.application.Variables import decompose_variable_value
from pyaedt.application.analysis_hf import ScatteringMethods
from pyaedt.generic.constants import AEDT_UNITS
from pyaedt.generic.constants import unit_converter
from pyaedt.generic.general_methods import check_and_download_folder
from pyaedt.generic.general_methods import conversion_function
from pyaedt.generic.general_methods import is_ironpython
from pyaedt.generic.general_methods import open_file
from pyaedt.generic.general_methods import pyaedt_function_handler
from pyaedt.generic.plot import ModelPlotter
from pyaedt.generic.plot import get_structured_mesh
from pyaedt.generic.plot import is_notebook
from pyaedt.generic.plot import plot_2d_chart
from pyaedt.generic.plot import plot_3d_chart
from pyaedt.generic.plot import plot_contour
from pyaedt.generic.plot import plot_polar_chart
from pyaedt.generic.settings import settings
from pyaedt.generic.touchstone_parser import read_touchstone

np = None
pv = None
if not is_ironpython:
    try:
        import numpy as np
    except ImportError:  # pragma: no cover
        np = None
    try:
        import pyvista as pv
    except ImportError:  # pragma: no cover
        pv = None


class FfdSolutionData(object):
    """Provides antenna far-field data.

    Read element pattern information in a JSON file generated by :func:`FfdSolutionDataExporter` and return the
    Python interface to plot and analyze the far-field data.

    Parameters
    ----------
    input_file : str
        Metadata information in a JSON file.
    frequency : float, optional
        Active frequency in hertz (Hz). The default is ``None``, in which case the first frequency is active.
    variation : str, optional
            Label to identify corresponding variation.
    model_info : dict, optional
    incident_power : dict, optional
        Dictionary with information of the incident power for each frequency.
        The default is ``None``, in which case an empty dictionary is applied.
        From AEDT 2024.1, this information is available from the XML input file.
        For example, the dictionary format for a two element farfield
        data = incident_power["1GHz"]
        data = [1, 0.99]
    touchstone_file : str, optional
        Touchstone file name. The default is ``None``.

    Examples
    --------

    >>> import pyaedt
    >>> from pyaedt.generic.farfield_visualization import FfdSolutionData
    >>> app = pyaedt.Hfss(version="2023.2", design="Antenna")
    >>> data = app.get_antenna_data()
    >>> metadata_file = data.metadata_file
    >>> app.release_desktop()
    >>> farfield_data = FfdSolutionData(input_file=metadata_file)
    >>> farfield_data.plot_3d(quantity_format="dB10")
    """

    def __init__(
        self, input_file, frequency=None, variation=None, model_info=None, incident_power=None, touchstone_file=None
    ):

        input_file_format = os.path.basename(input_file).split(".")[1]

        # Public
        self.output_dir = os.path.dirname(input_file)

        if input_file_format in ["txt", "xml"]:
            if not variation:
                variation = ""
            if not model_info:
                model_info = {}
            if not incident_power:
                incident_power = {}
            if not touchstone_file:
                touchstone_file = ""

            input_file = FfdSolutionDataExporter.export_pyaedt_antenna_metadata(
                input_file=input_file,
                output_dir=self.output_dir,
                variation=variation,
                model_info=model_info,
                power=incident_power,
                touchstone_file=touchstone_file,
            )

        # Protected
        self._mesh = None

        if not os.path.isfile(input_file):  # pragma: no cover
            raise Exception("JSON file does not exist.")

        # Private
        self.__logger = logger
        self.__input_file = input_file
        self.__raw_data = {}
        self.__freq_index = 0
        self.__model_units = "meter"

        self.__element_info = {}
        self.__frequencies = []
        self.__all_element_names = []
        self.__phase = {}
        self.__magnitude = {}
        self.__origin = []
        self.__taper = None
        self.__model_info = {}
        self.__is_array = False
        self.__component_objects = None
        self.__array_dimension = None
        self.__cell_position = None
        self.__lattice_vector = None
        self.__a_min = sys.maxsize
        self.__a_max = 0
        self.__b_min = sys.maxsize
        self.__b_max = 0
        self.__touchstone_data = None
        self.__weight = {}
        self.__phi_scan = 0.0
        self.__theta_scan = 0.0

        self.__incident_power_element = {}

        with open_file(input_file) as f:
            self.__metadata = json.load(f)

        if not self.metadata:  # pragma: no cover
            raise Exception("Metadata could not be loaded..")

        elements = self.metadata["element_pattern"]
        for element_name, element_props in elements.items():
            location = element_props["location"]
            pattern_file = os.path.join(self.output_dir, element_props["file_name"])
            incident_power = element_props["incident_power"]
            accepted_power = element_props["accepted_power"]
            radiated_power = element_props["radiated_power"]

            new_incident_power = {}

            if incident_power:
                for power_freq in incident_power:
                    value = incident_power[power_freq]
                    frequency = power_freq
                    if isinstance(power_freq, str):
                        frequency, units = decompose_variable_value(power_freq)
                        if units:  # pragma: no cover
                            frequency = unit_converter(frequency, "Freq", units, "Hz")
                    new_incident_power[frequency] = value

            new_radiated_power = {}
            if radiated_power:
                for power_freq in radiated_power:
                    frequency = power_freq
                    if isinstance(power_freq, str):
                        frequency, units = decompose_variable_value(power_freq)
                        if units:  # pragma: no cover
                            frequency = unit_converter(frequency, "Freq", units, "Hz")
                    new_radiated_power[frequency] = radiated_power[power_freq]

            new_accepted_power = {}
            if accepted_power:
                for power_freq in accepted_power:
                    frequency = power_freq
                    if isinstance(power_freq, str):
                        frequency, units = decompose_variable_value(power_freq)
                        if units:  # pragma: no cover
                            frequency = unit_converter(frequency, "Freq", units, "Hz")
                    new_accepted_power[frequency] = accepted_power[power_freq]

            self.__element_info[element_name] = {
                "pattern_file": pattern_file,
                "location": [float(location[0]), float(location[1]), float(location[2])],
                "incident_power": new_incident_power,
                "accepted_power": new_accepted_power,
                "radiated_power": new_radiated_power,
            }

        if not self.element_info:  # pragma: no cover
            raise Exception("Wrong farfield file load.")

        # Update properties with the loaded information
        self.__all_element_names = list(self.element_info.keys())
        for element in self.all_element_names:
            self.__magnitude[element] = 1.0
            self.__phase[element] = 0.0

        self.__origin = [0, 0, 0]
        self.__taper = "flat"

        # Load farfield data
        is_farfield_loaded = self.__init_ffd(self.element_info)
        if not is_farfield_loaded:  # pragma: no cover
            raise Exception("Farfield information from ffd files can not be loaded.")

        # Load touchstone data
        metadata_touchstone = os.path.join(self.output_dir, self.metadata.get("touchstone_file", None))

        if not touchstone_file:
            touchstone_file = metadata_touchstone

        if touchstone_file and os.path.isfile(touchstone_file):
            self.__touchstone_data = read_touchstone(touchstone_file)

        required_array_keys = ["array_dimension", "component_objects", "lattice_vector", "cell_position"]
        if all(key in self.metadata for key in required_array_keys):
            self.__is_array = True
            self.__component_objects = self.metadata["component_objects"]
            self.__array_dimension = self.metadata["array_dimension"]
            self.__cell_position = self.metadata["cell_position"]
            self.__lattice_vector = self.metadata["lattice_vector"]
        else:
            self.__is_array = False

        # Get element indices
        port_indices = self.get_port_index()
        if not port_indices:  # pragma: no cover
            raise Exception("Wrong port index load.")

        # Update active frequency if passed in the initialization
        if frequency and frequency in self.frequencies:
            freq_index = self.frequencies.index(frequency)
            self.frequency = self.frequencies[freq_index]
        else:
            self.frequency = self.frequencies[0]

        port_indices_array = np.array(list(port_indices.values()))

        rows = port_indices_array[:, 0] - 1
        cols = port_indices_array[:, 1] - 1

        self.__a_min = min(self.__a_min, np.min(rows))
        self.__a_max = max(self.__a_max, np.max(rows))
        self.__b_min = min(self.__b_min, np.min(cols))
        self.__b_max = max(self.__b_max, np.max(cols))

    @property
    def phi_scan(self):
        """Phi scan angle in degrees. It applies only for arrays."""
        return self.__phi_scan

    @phi_scan.setter
    def phi_scan(self, value):
        self.__phi_scan = value
        self.__element_weight()

    @property
    def theta_scan(self):
        """Theta scan angle in degrees. It applies only for arrays."""
        return self.__theta_scan

    @theta_scan.setter
    def theta_scan(self, value):
        self.__theta_scan = value
        self.__element_weight()

    @property
    def metadata(self):
        """Antenna metadata."""
        return self.__metadata

    @property
    def touchstone_data(self):
        """Touchstone data."""
        return self.__touchstone_data

    @property
    def s_parameters(self):
        """Passive s-parameters."""
        if self.touchstone_data:
            touchstone_frequencies = self.touchstone_data.f
            index = np.abs(touchstone_frequencies - self.frequency).argmin()
            return self.touchstone_data.s[index]

    @property
    def incident_power_element(self):
        """Incident power per element in watts."""
        incident_power = {}
        for element_name, element_props in self.element_info.items():
            element_power = element_props.get("incident_power", None)
            if element_power and element_power.get(self.frequency, None):
                incident_power[element_name] = element_power.get(self.frequency) * self.magnitude[element_name]
            else:  # pragma: no cover
                incident_power[element_name] = 1.0 * self.magnitude[element_name]
        self.__incident_power_element = incident_power
        return self.__incident_power_element

    @property
    def incident_power(self):
        """Total incident power in watts."""
        incident_power_element = self.incident_power_element
        if incident_power_element:
            return sum(incident_power_element.values())

    @property
    def accepted_power_element(self):
        """Accepted power per element in watts."""
        power = {}
        for element_name, element_props in self.element_info.items():
            element_power = element_props.get("accepted_power", None)
            if element_power and element_power.get(self.frequency, None):
                power[element_name] = element_power.get(self.frequency) * self.magnitude[element_name]
            else:  # pragma: no cover
                power[element_name] = 1.0 * self.magnitude[element_name]
        return power

    @property
    def accepted_power(self):
        """Total accepted power in watts."""
        power_element = self.accepted_power_element
        if power_element:
            return sum(power_element.values())

    @property
    def radiated_power_element(self):
        """Radiated power per element in watts."""
        power = {}
        for element_name, element_props in self.element_info.items():
            element_power = element_props.get("radiated_power", None)
            if element_power and element_power.get(self.frequency, None):
                power[element_name] = element_power.get(self.frequency) * self.magnitude[element_name]
            else:  # pragma: no cover
                power[element_name] = 1.0 * self.magnitude[element_name]
        return power

    @property
    def radiated_power(self):
        """Total radiated power in watts."""
        power_element = self.radiated_power_element
        if power_element:
            return sum(power_element.values())

    @property
    def active_s_parameters(self):
        """Active s-parameters."""
        if self.s_parameters is not None:
            active_s_parameter = {}
            incident_power_list = list(self.incident_power_element.values())
            phase_list = list(self.phase.values())

            for element_cont, element in enumerate(self.all_element_names):
                row_s_parameters = self.s_parameters[element_cont]
                active_s_parameter[element] = 0

                active_amplitude_mag = np.sqrt(incident_power_list[element_cont])
                active_amplitude_phase = phase_list[element_cont]
                active_amplitude_real = active_amplitude_mag * np.cos(active_amplitude_phase)
                active_amplitude_imag = active_amplitude_mag * np.sin(active_amplitude_phase)
                active_amplitude = active_amplitude_real + 1j * active_amplitude_imag

                for s_param_cont, s_parameter_value in enumerate(row_s_parameters):
                    amplitude_mag = np.sqrt(incident_power_list[s_param_cont])
                    amplitude_phase = phase_list[s_param_cont]
                    amplitude_real = amplitude_mag * np.cos(amplitude_phase)
                    amplitude_imag = amplitude_mag * np.sin(amplitude_phase)
                    amplitude = amplitude_real + 1j * amplitude_imag

                    if active_amplitude == 0:
                        active_s_parameter[element] = None
                    else:
                        active_s_parameter[element] += s_parameter_value * amplitude / active_amplitude
            return active_s_parameter

    @property
    def input_file(self):
        """Input file."""
        return self.__input_file

    @property
    def farfield_data(self):
        """Farfield data."""
        return self.combine_farfield(self.theta_scan, self.phi_scan)

    @property
    def element_info(self):
        """File information."""
        return self.__element_info

    @property
    def frequencies(self):
        """Available frequencies."""
        return self.__frequencies

    @property
    def all_element_names(self):
        """Available port names."""
        return self.__all_element_names

    @property
    def weight(self):
        """Weight."""
        return self.__weight

    @property
    def frequency(self):
        """Active frequency."""
        return self._frequency

    @frequency.setter
    def frequency(self, val):
        if isinstance(val, str):
            frequency, units = decompose_variable_value(val)
            unit_converter(frequency, "Freq", units, "Hz")
            val = frequency
        if val in self.frequencies:
            self._frequency = val
            self.__freq_index = self.frequencies.index(val)
        else:  # pragma: no cover
            self.__logger.error("Frequency not available.")

    @property
    def phase(self):
        """Phase offset in degrees on each port."""
        return self.__phase

    @phase.setter
    def phase(self, phases):
        if len(phases) != len(self.all_element_names):
            self.__logger.error("Number of phases must be equal to number of ports.")
        else:
            self.__phase = phases

    @property
    def magnitude(self):
        """Magnitude weight applied on each port."""
        return self.__magnitude

    @magnitude.setter
    def magnitude(self, mags):
        if len(mags) != len(self.all_element_names):
            self.__logger.error("Number of magnitude values must be equal to number of ports.")
        else:
            self.__magnitude = mags

    @property
    def taper(self):
        """Taper type.

        Options are:

        - ``"cosine"``
        - ``"flat"``
        - ``"hamming"``
        - ``"triangular"``
        - ``"uniform"``
        """
        return self.__taper

    @taper.setter
    def taper(self, val):
        if val.lower() in ("flat", "uniform", "cosine", "triangular", "hamming"):
            self.__taper = val

        else:
            self.__logger.error("This taper is not implemented")

    @property
    def origin(self):
        """Far field origin in meters."""
        return self.__origin

    @origin.setter
    def origin(self, vals):
        if len(vals) != 3:
            self.__logger.error("Origin is wrong.")
        else:
            self.__origin = vals

    @pyaedt_function_handler()
    def combine_farfield(self, phi_scan=0.0, theta_scan=0.0):
        """Compute the far field pattern calculated for a specific phi and theta scan angle requested.

        Parameters
        ----------
        phi_scan : float, optional
            Phi scan angle in degrees. The default is ``0.0``.
        theta_scan : float, optional
            Theta scan angle in degrees. The default is ``0.0``.

        Returns
        -------
        dict
            Far field data dictionary.
        """
        # Modify theta and phi and compute weight

        self.__theta_scan = theta_scan
        self.__phi_scan = phi_scan
        self.__element_weight()

        port_positions = {}

        for port_name in self.all_element_names:
            port_positions[port_name] = self.element_info[port_name]["location"]

        # Combine farfield of each port
        initial_port = self.all_element_names[0]
        freq_name_key = self.frequencies[self.__freq_index]
        data = self.__raw_data[initial_port][freq_name_key]
        length_of_ff_data = len(data["rETheta"])
        theta_range = data["Theta"]
        phi_range = data["Phi"]
        n_theta = len(theta_range)
        n_phi = len(phi_range)

        incident_power = self.incident_power
        radiated_power = self.radiated_power
        accepted_power = self.accepted_power

        ph, th = np.meshgrid(data["Phi"], data["Theta"])
        ph = np.deg2rad(ph)
        th = np.deg2rad(th)
        c = 299792458
        k = 2 * np.pi * self.frequency / c
        kx_grid = k * np.sin(th) * np.cos(ph)
        ky_grid = k * np.sin(th) * np.sin(ph)
        kz_grid = k * np.cos(th)
        kx_flat = kx_grid.ravel()
        ky_flat = ky_grid.ravel()
        kz_flat = kz_grid.ravel()
        rEphi_fields_sum = np.zeros(length_of_ff_data, dtype=complex)
        rETheta_fields_sum = np.zeros(length_of_ff_data, dtype=complex)

        # Farfield superposition
        for _, port in enumerate(self.all_element_names):
            data = self.__raw_data[port][freq_name_key]
            if port not in self.weight:  # pragma: no cover
                self.__weight[port] = np.sqrt(0) * np.exp(1j * 0)

            xyz_pos = port_positions[port]
            weight = self.weight[port]
            array_factor = np.exp(1j * (xyz_pos[0] * kx_flat + xyz_pos[1] * ky_flat + xyz_pos[2] * kz_flat)) * weight
            rEphi_fields_sum += array_factor * data["rEPhi"]
            rETheta_fields_sum += array_factor * data["rETheta"]

        # Farfield origin shift
        origin = self.origin
        array_factor = np.exp(-1j * (origin[0] * kx_flat + origin[1] * ky_flat + origin[2] * kz_flat))
        rETheta_fields_sum = array_factor * rETheta_fields_sum
        rEphi_fields_sum = array_factor * rEphi_fields_sum

        rEtheta_fields_sum = np.reshape(rETheta_fields_sum, (n_theta, n_phi))
        rEphi_fields_sum = np.reshape(rEphi_fields_sum, (n_theta, n_phi))

        farfield_data = OrderedDict()
        farfield_data["rEPhi"] = rEphi_fields_sum
        farfield_data["rETheta"] = rEtheta_fields_sum
        farfield_data["rETotal"] = np.sqrt(
            np.power(np.abs(rEphi_fields_sum), 2) + np.power(np.abs(rEtheta_fields_sum), 2)
        )
        farfield_data["Theta"] = theta_range
        farfield_data["Phi"] = phi_range
        farfield_data["nPhi"] = n_phi
        farfield_data["nTheta"] = n_theta

        real_gain = 2 * np.pi * np.abs(np.power(farfield_data["rETotal"], 2)) / incident_power / 377
        farfield_data["RealizedGain"] = real_gain
        farfield_data["RealizedGain_Total"] = real_gain
        farfield_data["RealizedGain_dB"] = 10 * np.log10(real_gain)
        real_gain = 2 * np.pi * np.abs(np.power(farfield_data["rETheta"], 2)) / incident_power / 377
        farfield_data["RealizedGain_Theta"] = real_gain
        real_gain = 2 * np.pi * np.abs(np.power(farfield_data["rEPhi"], 2)) / incident_power / 377
        farfield_data["RealizedGain_Phi"] = real_gain

        gain = 2 * np.pi * np.abs(np.power(farfield_data["rETotal"], 2)) / accepted_power / 377
        farfield_data["Gain"] = gain
        farfield_data["Gain_Total"] = gain
        farfield_data["Gain_dB"] = 10 * np.log10(gain)
        gain = 2 * np.pi * np.abs(np.power(farfield_data["rETheta"], 2)) / accepted_power / 377
        farfield_data["Gain_Theta"] = gain
        gain = 2 * np.pi * np.abs(np.power(farfield_data["rEPhi"], 2)) / accepted_power / 377
        farfield_data["Gain_Phi"] = gain

        directivity = 2 * np.pi * np.abs(np.power(farfield_data["rETotal"], 2)) / radiated_power / 377
        farfield_data["Directivity"] = directivity
        farfield_data["Directivity_Total"] = directivity
        farfield_data["Directivity_dB"] = 10 * np.log10(directivity)
        directivity = 2 * np.pi * np.abs(np.power(farfield_data["rETheta"], 2)) / radiated_power / 377
        farfield_data["Directivity_Theta"] = directivity
        directivity = 2 * np.pi * np.abs(np.power(farfield_data["rEPhi"], 2)) / radiated_power / 377
        farfield_data["Directivity_Phi"] = directivity

        # farfield_data["Element_Location"] = port_positions
        return farfield_data

    @pyaedt_function_handler()
    def get_accepted_power(self):
        """Compute the accepted power from active s-parameters and incident power.

        Returns
        -------
        float
            Total accepted power.
        """

        if self.active_s_parameters is not None:
            accepted_power = {}
            for _, element in enumerate(self.all_element_names):
                if self.active_s_parameters[element] is not None:
                    operation = 1 - np.power(np.abs(self.active_s_parameters[element]), 2)
                    accepted_power[element] = self.incident_power_element[element] * operation
                else:
                    accepted_power[element] = 0.0
            total_accepted_power = sum(accepted_power.values())
            return total_accepted_power

    @pyaedt_function_handler()
    def __assign_weight_taper(self, a, b):  # pragma: no cover
        """Assign weight to array.

        Parameters
        ----------
        a : int
            Index of array, column.
        b : int
            Index of array, row.

        Returns
        -------
        float
            Weight applied to specific index of the array.
        """
        taper = self.taper

        if taper.lower() in ("flat", "uniform") or not self.__is_array:
            return 1

        cosinePow = 1
        edgeTaper_dB = -200

        edgeTaper = 10 ** ((float(edgeTaper_dB)) / 20)

        threshold = 1e-10

        # find the distance between current cell and array center in terms of index
        lattice_vector = self.__lattice_vector[self.__freq_index]
        if not lattice_vector or not len(lattice_vector) == 6:
            return 1.0

        center_a = (self.__a_min + self.__a_max) / 2
        center_b = (self.__b_min + self.__b_max) / 2

        length_in_direction1 = a - center_a
        length_in_direction2 = b - center_b
        max_length_in_dir1 = self.__a_max - self.__a_min
        max_length_in_dir2 = self.__b_max - self.__b_min

        if taper.lower() == "cosine":  # Cosine
            if max_length_in_dir1 < threshold:
                w1 = 1
            else:
                w1 = (1 - edgeTaper) * (
                    math.cos(math.pi * length_in_direction1 / max_length_in_dir1)
                ) ** cosinePow + edgeTaper
            if max_length_in_dir2 < threshold:
                w2 = 1
            else:
                w2 = (1 - edgeTaper) * (
                    math.cos(math.pi * length_in_direction2 / max_length_in_dir2)
                ) ** cosinePow + edgeTaper
        elif taper.lower() == "triangular":  # Triangular
            if max_length_in_dir1 < threshold:
                w1 = 1
            else:
                w1 = (1 - edgeTaper) * (1 - (math.fabs(length_in_direction1) / (max_length_in_dir1 / 2))) + edgeTaper
            if max_length_in_dir2 < threshold:
                w2 = 1
            else:
                w2 = (1 - edgeTaper) * (1 - (math.fabs(length_in_direction2) / (max_length_in_dir2 / 2))) + edgeTaper
        elif taper.lower() == "hamming":  # Hamming Window
            if max_length_in_dir1 < threshold:
                w1 = 1
            else:
                w1 = 0.54 - 0.46 * math.cos(2 * math.pi * (length_in_direction1 / max_length_in_dir1 - 0.5))
            if max_length_in_dir2 < threshold:
                w2 = 1
            else:
                w2 = 0.54 - 0.46 * math.cos(2 * math.pi * (length_in_direction2 / max_length_in_dir2 - 0.5))
        else:
            return 0

        return w1 * w2

    @pyaedt_function_handler()
    def __phase_shift_steering(self, a, b, theta=0.0, phi=0.0):
        """Shift element phase for a specific Theta and Phi scan angle in degrees.

        This method calculates phase shifts between array elements in A and B directions given the lattice vector.

        Parameters
        ----------
        a : int
            Index of array, column.
        b : int
            Index of array, row.
        theta : float, optional
            Theta scan angle in degrees. The default is ``0.0``.
        phi : float, optional
            Phi scan angle in degrees. The default is ``0.0``.

        Returns
        -------
        float
            Phase shift in degrees.
        """
        c = 299792458
        k = (2 * math.pi * self.frequency) / c
        a = int(a)
        b = int(b)
        theta = np.deg2rad(theta)
        phi = np.deg2rad(phi)

        lattice_vector = self.__lattice_vector

        a_x, a_y, b_x, b_y = [lattice_vector[0], lattice_vector[1], lattice_vector[3], lattice_vector[4]]

        phase_shift_a = -((a_x * k * np.sin(theta) * np.cos(phi)) + (a_y * k * np.sin(theta) * np.sin(phi)))

        phase_shift_b = -((b_x * k * np.sin(theta) * np.cos(phi)) + (b_y * k * np.sin(theta) * np.sin(phi)))

        phase_shift = a * phase_shift_a + b * phase_shift_b

        return np.rad2deg(phase_shift)

    @pyaedt_function_handler()
    def __element_weight(self):
        # Obtain weights for each element
        for element_name in self.all_element_names:
            amplitude = self.magnitude[element_name]
            phase = self.phase[element_name]
            if self.__is_array:
                index_port = self.get_port_index()
                index_str = index_port[element_name]
                a = index_str[0] - 1
                b = index_str[1] - 1
                phase_steering = self.__phase_shift_steering(a, b, self.theta_scan, self.phi_scan)
                phase += phase_steering
                amplitude_taper = self.__assign_weight_taper(a=a, b=b)
                amplitude *= amplitude_taper

            self.__weight[element_name] = np.sqrt(amplitude) * np.exp(1j * np.deg2rad(phase))
            self.__magnitude[element_name] = amplitude
            self.__phase[element_name] = phase

    @pyaedt_function_handler()
    def plot_contour(
        self,
        quantity="RealizedGain",
        phi=0,
        theta=0,
        title=None,
        quantity_format="dB10",
        output_file=None,
        levels=64,
        polar=True,
        max_theta=180,
        show=True,
    ):
        """Create a contour plot of a specified quantity in Matplotlib.

        Parameters
        ----------
        quantity : str, optional
            Far field quantity to plot. The default is ``"RealizedGain"``.
            Available quantities are: ``"RealizedGain"``, ``"RealizedGain_Phi"``, ``"RealizedGain_Theta"``,
            ``"rEPhi"``, ``"rETheta"``, and ``"rETotal"``.
        phi : float, int, optional
            Phi scan angle in degrees. The default is ``0``.
        theta : float, int, optional
            Theta scan angle in degrees. The default is ``0``.
        title : str, optional
            Plot title. The default is ``"RectangularPlot"``.
        quantity_format : str, optional
            Conversion data function.
            Available functions are: ``"abs"``, ``"ang"``, ``"dB10"``, ``"dB20"``, ``"deg"``, ``"imag"``, ``"norm"``,
            and ``"real"``.
        output_file : str, optional
            Full path for the image file. The default is ``None``, in which case the file is not exported.
        levels : int, optional
            Color map levels. The default is ``64``.
        show : bool, optional
            Whether to show the plot. The default is ``True``. If ``False``, the Matplotlib
            instance of the plot is shown.
        polar : bool, optional
            Generate the plot in polar coordinates. The default is ``True``. If ``False``, the plot
            generated is rectangular.
        max_theta : float or int, optional
            Maximum theta angle for plotting. The default is ``180``, which plots the far-field data for
            all angles. Setting ``max_theta`` to 90 limits the displayed data to the upper
            hemisphere, that is (0 < theta < 90).

        Returns
        -------
        :class:`matplotlib.pyplot.Figure`
            Matplotlib figure object.

        Examples
        --------
        >>> import pyaedt
        >>> app = pyaedt.Hfss(version="2024.2", design="Antenna")
        >>> setup_name = "Setup1 : LastAdaptive"
        >>> frequencies = [77e9]
        >>> sphere = "3D"
        >>> data = app.get_antenna_data(frequencies,setup_name,sphere)
        >>> data.plot_contour()

        """
        if not title:
            title = quantity

        data = self.combine_farfield(phi, theta)
        if quantity not in data:  # pragma: no cover
            raise Exception("Far field quantity is not available.")

        select = np.abs(data["Theta"]) <= max_theta  # Limit theta range for plotting.

        data_to_plot = data[quantity][select, :]
        data_to_plot = conversion_function(data_to_plot, quantity_format)
        if not isinstance(data_to_plot, np.ndarray):  # pragma: no cover
            raise Exception("Wrong format quantity.")

        ph, th = np.meshgrid(data["Phi"], data["Theta"][select])
        # Convert to radians for polar plot.
        ph = np.radians(ph) if polar else ph

        return plot_contour(
            plot_data=[data_to_plot, th, ph],
            xlabel=r"$\phi$ (Degrees)",
            ylabel=r"$\theta$ (Degrees)",
            title=title,
            levels=levels,
            polar=polar,
            snapshot_path=output_file,
            max_theta=max_theta,
            color_bar=quantity_format,
            show=show,
        )

    @pyaedt_function_handler()
    def plot_cut(
        self,
        quantity="RealizedGain",
        primary_sweep="phi",
        secondary_sweep_value=0,
        phi=0,
        theta=0,
        title="Far Field Cut",
        quantity_format="dB10",
        output_file=None,
        show=True,
        is_polar=False,
        show_legend=True,
    ):
        """Create a 2D plot of a specified quantity in Matplotlib.

        Parameters
        ----------
        quantity : str, optional
            Quantity to plot. The default is ``"RealizedGain"``.
            Available quantities are: ``"RealizedGain"``, ``"RealizedGain_Theta"``, ``"RealizedGain_Phi"``,
            ``"rETotal"``, ``"rETheta"``, and ``"rEPhi"``.
        primary_sweep : str, optional.
            X-axis variable. The default is ``"phi"``. Options are ``"phi"`` and ``"theta"``.
        secondary_sweep_value : float, list, string, optional
            List of cuts on the secondary sweep to plot. The default is ``0``. Options are
            `"all"`, a single value float, or a list of float values.
        phi : float, int, optional
            Phi scan angle in degrees. The default is ``0``.
        theta : float, int, optional
            Theta scan angle in degrees. The default is ``0``.
        title : str, optional
            Plot title. The default is ``"RectangularPlot"``.
        quantity_format : str, optional
            Conversion data function.
            Available functions are: ``"abs"``, ``"ang"``, ``"dB10"``, ``"dB20"``, ``"deg"``, ``"imag"``, ``"norm"``,
            and ``"real"``.
        output_file : str, optional
            Full path for the image file. The default is ``None``, in which case an image in not exported.
        show : bool, optional
            Whether to show the plot. The default is ``True``.
            If ``False``, the Matplotlib instance of the plot is shown.
        is_polar : bool, optional
            Whether this plot is a polar plot. The default is ``True``.
        show_legend : bool, optional
            Whether to display the legend or not. The default is ``True``.

        Returns
        -------
        :class:`matplotlib.pyplot.Figure`
            Matplotlib figure object.
            If ``show=True``, a Matplotlib figure instance of the plot is returned.
            If ``show=False``, the plotted curve is returned.

        Examples
        --------
        >>> import pyaedt
        >>> app = pyaedt.Hfss(version="2023.2", design="Antenna")
        >>> setup_name = "Setup1 : LastAdaptive"
        >>> frequencies = [77e9]
        >>> sphere = "3D"
        >>> data = app.get_antenna_data(frequencies,setup_name,sphere)
        >>> data.plot_cut(theta=20)
        """

        data = self.combine_farfield(phi, theta)
        if quantity not in data:  # pragma: no cover
            raise Exception("Far field quantity not available.")

        data_to_plot = data[quantity]

        curves = []
        if primary_sweep.lower() == "phi":
            x_key, y_key = "Phi", "Theta"
            temp = data_to_plot
        else:
            y_key, x_key = "Phi", "Theta"
            temp = data_to_plot.T
        x = data[x_key]
        if is_polar:
            x = [i * 2 * math.pi / 360 for i in x]
        if secondary_sweep_value == "all":
            for el in data[y_key]:
                idx = self.__find_nearest(data[y_key], el)
                y = temp[idx]
                y = conversion_function(y, quantity_format)
                if not isinstance(y, np.ndarray):  # pragma: no cover
                    raise Exception("Format of quantity is wrong.")
                curves.append([x, y, "{}={}".format(y_key, el)])
        elif isinstance(secondary_sweep_value, list):
            list_inserted = []
            for el in secondary_sweep_value:
                theta_idx = self.__find_nearest(data[y_key], el)
                if theta_idx not in list_inserted:
                    y = temp[theta_idx]
                    y = conversion_function(y, quantity_format)
                    if not isinstance(y, np.ndarray):  # pragma: no cover
                        raise Exception("Format of quantity is wrong.")
                    curves.append([x, y, "{}={}".format(y_key, el)])
                    list_inserted.append(theta_idx)
        else:
            theta_idx = self.__find_nearest(data[y_key], secondary_sweep_value)
            y = temp[theta_idx]
            y = conversion_function(y, quantity_format)
            if not isinstance(y, np.ndarray):  # pragma: no cover
                raise Exception("Wrong format quantity.")
            curves.append([x, y, "{}={}".format(y_key, data[y_key][theta_idx])])

        if is_polar:
            return plot_polar_chart(
                curves,
                xlabel=x_key,
                ylabel=quantity,
                title=title,
                snapshot_path=output_file,
                show_legend=show_legend,
                show=show,
            )
        else:
            return plot_2d_chart(
                curves,
                xlabel=x_key,
                ylabel=quantity,
                title=title,
                snapshot_path=output_file,
                show_legend=show_legend,
                show=show,
            )

    @pyaedt_function_handler()
    def plot_3d_chart(
        self,
        quantity="RealizedGain",
        phi=0,
        theta=0,
        title="3D Plot",
        quantity_format="dB10",
        output_file=None,
        show=True,
    ):
        """Create a 3D chart of a specified quantity in Matplotlib.

        Parameters
        ----------
        quantity : str, optional
            Far field quantity to plot. The default is ``"RealizedGain"``.
            Available quantities are: ``"RealizedGain"``, ``"RealizedGain_Phi"``, ``"RealizedGain_Theta"``,
            ``"rEPhi"``, ``"rETheta"``, and ``"rETotal"``.
        phi : float, int, optional
            Phi scan angle in degree. The default is ``0``.
        theta : float, int, optional
            Theta scan angle in degree. The default is ``0``.
        title : str, optional
            Plot title. The default is ``"3D Plot"``.
        quantity_format : str, optional
            Conversion data function.
            Available functions are: ``"abs"``, ``"ang"``, ``"dB10"``, ``"dB20"``, ``"deg"``, ``"imag"``, ``"norm"``,
            and ``"real"``.
        output_file : str, optional
            Full path for the image file. The default is ``None``, in which case a file is not exported.
        show : bool, optional
            Whether to show the plot. The default is ``True``.
            If ``False``, the Matplotlib instance of the plot is not shown.

        Returns
        -------
        :class:`matplotlib.pyplot.Figure`
            Matplotlib figure object.
            If ``show=True``, a Matplotlib figure instance of the plot is returned.
            If ``show=False``, the plotted curve is returned.

        Examples
        --------
        >>> import pyaedt
        >>> app = pyaedt.Hfss(version="2023.2", design="Antenna")
        >>> setup_name = "Setup1 : LastAdaptive"
        >>> frequencies = [77e9]
        >>> sphere = "3D"
        >>> data = app.get_antenna_data(frequencies,setup_name,sphere)
        >>> data.polar_plot_3d(theta=10)
        """
        data = self.combine_farfield(phi, theta)
        if quantity not in data:  # pragma: no cover
            raise Exception("Far field quantity is not available.")

        ff_data = conversion_function(data[quantity], quantity_format)
        if not isinstance(ff_data, np.ndarray):  # pragma: no cover
            raise Exception("Format of the quantity is wrong.")

        # re-normalize to 0 and 1
        ff_max = np.max(ff_data)
        ff_min = np.min(ff_data)

        ff_data_renorm = (ff_data - ff_min) / (ff_max - ff_min)

        theta = np.deg2rad(np.array(data["Theta"]))
        phi = np.deg2rad(np.array(data["Phi"]))
        phi_grid, theta_grid = np.meshgrid(phi, theta)
        r = np.reshape(ff_data_renorm, (len(data["Theta"]), len(data["Phi"])))

        x = r * np.sin(theta_grid) * np.cos(phi_grid)
        y = r * np.sin(theta_grid) * np.sin(phi_grid)
        z = r * np.cos(theta_grid)
        return plot_3d_chart([x, y, z], xlabel="Theta", ylabel="Phi", title=title, snapshot_path=output_file, show=show)

    @pyaedt_function_handler()
    def plot_3d(
        self,
        quantity="RealizedGain",
        quantity_format="dB10",
        rotation=None,
        output_file=None,
        show=True,
        show_as_standalone=False,
        pyvista_object=None,
        background=None,
        scale_farfield=None,
        show_beam_slider=True,
        show_geometry=True,
    ):
        """Create a 3D polar plot of the geometry with a radiation pattern in PyVista.

        Parameters
        ----------
        quantity : str, optional
            Quantity to plot. The default is ``"RealizedGain"``.
            Available quantities are: ``"RealizedGain"``, ``"RealizedGain_Theta"``, ``"RealizedGain_Phi"``,
            ``"rETotal"``, ``"rETheta"``, and ``"rEPhi"``.
        quantity_format : str, optional
            Conversion data function.
            Available functions are: ``"abs"``, ``"ang"``, ``"dB10"``, ``"dB20"``, ``"deg"``, ``"imag"``, ``"norm"``,
            and ``"real"``.
        output_file : str, optional
            Full path for the image file. The default is ``None``, in which case a file is not exported.
        rotation : list, optional
            Far field rotation matrix. The matrix contains three vectors, around x, y, and z axes.
            The default is ``[[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]]``.
        show : bool, optional
            Whether to show the plot. The default is ``True``.
        show_as_standalone : bool, optional
            Whether to show a plot as standalone. The default is ``False``.
        pyvista_object : :class:`Pyvista.Plotter`, optional
            PyVista instance defined externally. The default is ``None``.
        background : list or str, optional
            Background color if a list is passed or background picture if a string is passed.
            The default is ``None``.
        scale_farfield : list, optional
            List with minimum and maximum values of the scale slider. The default is
            ``None``.
        show_beam_slider : bool, optional
            Whether the Theta and Phi scan slider is active. The default is ``True``.
        show_geometry :
            Whether to show the geometry. The default is ``True``.

        Returns
        -------
        bool or :class:`Pyvista.Plotter`
            ``True`` when successful. The :class:`Pyvista.Plotter` is returned when ``show`` and
            ``image_path`` are ``False``.

        Examples
        --------
        >>> import pyaedt
        >>> app = pyaedt.Hfss(version="2023.2", design="Antenna")
        >>> setup_name = "Setup1 : LastAdaptive"
        >>> frequencies = [77e9]
        >>> sphere = "3D"
        >>> data = app.get_antenna_data(setup=setup_name,sphere=sphere)
        >>> data.plot_3d(quantity_format="dB10")
        """
        if not rotation:
            rotation = np.eye(3)
        elif isinstance(rotation, (list, tuple)):  # pragma: no cover
            rotation = np.array(rotation)
        text_color = "white"
        if background is None:
            background = [255, 255, 255]
            text_color = "black"

        farfield_data = self.combine_farfield(phi_scan=0, theta_scan=0)
        if quantity not in farfield_data:  # pragma: no cover
            raise Exception("Far field quantity is not available.")

        self._mesh = self.get_far_field_mesh(quantity=quantity, quantity_format=quantity_format)

        rotation_euler = self.__rotation_to_euler_angles(rotation) * 180 / np.pi

        if not output_file and not show:
            off_screen = False
        else:
            off_screen = not show

        if not pyvista_object:
            if show_as_standalone:  # pragma: no cover
                p = pv.Plotter(notebook=False)
            else:
                is_notebook_flag = is_notebook()
                p = pv.Plotter(notebook=is_notebook_flag)
            p.off_screen = off_screen
            p.enable_ssao()
            p.enable_parallel_projection()
        else:  # pragma: no cover
            p = pyvista_object

        uf = UpdateBeamForm(self, quantity, quantity_format)

        default_background = [255, 255, 255]
        axes_color = [i / 255 for i in default_background]

        if isinstance(background, list):
            background_color = [i / 255 for i in background]
            p.background_color = background_color
            axes_color = [0 if i >= 128 else 255 for i in background]
        elif isinstance(background, str):  # pragma: no cover
            p.add_background_image(background, scale=2.5)

        if show_beam_slider and self.__is_array:
            p.add_slider_widget(
                uf.update_phi,
                rng=[0, 360],
                value=0,
                title="Phi",
                pointa=(0.55, 0.1),
                pointb=(0.74, 0.1),
                style="modern",
                interaction_event="always",
                title_height=0.02,
                color=axes_color,
            )
            p.add_slider_widget(
                uf.update_theta,
                rng=[-180, 180],
                value=0,
                title="Theta",
                pointa=(0.77, 0.1),
                pointb=(0.98, 0.1),
                style="modern",
                interaction_event="always",
                title_height=0.02,
                color=axes_color,
            )

        sargs = dict(
            title_font_size=12,
            label_font_size=12,
            shadow=True,
            n_labels=7,
            italic=True,
            fmt="%.1f",
            font_family="arial",
            vertical=True,
            position_x=0.05,
            position_y=0.65,
            height=0.3,
            width=0.06,
            color=axes_color,
            title=None,
            outline=False,
        )

        cad_mesh = self.__get_geometry(off_screen=off_screen)

        data = conversion_function(farfield_data[quantity], function=quantity_format)
        if not isinstance(data, np.ndarray):  # pragma: no cover
            raise Exception("Wrong format quantity.")

        max_data = np.max(data)
        min_data = np.min(data)
        ff_mesh_inst = p.add_mesh(uf.output, cmap="jet", clim=[min_data, max_data], scalar_bar_args=sargs)

        if cad_mesh:  # pragma: no cover

            def toggle_vis_ff(flag):
                ff_mesh_inst.SetVisibility(flag)

            def toggle_vis_cad(flag):
                for i in cad:
                    i.SetVisibility(flag)

            def scale(value=1):
                ff_mesh_inst.SetScale(value, value, value)
                sf = AEDT_UNITS["Length"][self.__model_units]
                ff_mesh_inst.SetPosition(np.divide(self.origin, sf))
                ff_mesh_inst.SetOrientation(rotation_euler)

            p.add_checkbox_button_widget(toggle_vis_ff, value=True, size=30)
            p.add_text("Show Far Fields", position=(70, 25), color=text_color, font_size=10)
            if not scale_farfield:
                if self.__is_array:
                    slider_max = int(np.ceil(np.abs(np.max(self.__array_dimension) / np.min(np.abs(p.bounds)))))
                else:  # pragma: no cover
                    slider_max = int(np.ceil((np.max(p.bounds) / 2 / np.min(np.abs(p.bounds)))))
                slider_min = 0
            else:
                slider_max = scale_farfield[1]
                slider_min = scale_farfield[0]
            value = slider_max / 3

            p.add_slider_widget(
                scale,
                [slider_min, slider_max],
                title="Scale Plot",
                value=value,
                pointa=(0.7, 0.93),
                pointb=(0.99, 0.93),
                style="modern",
                title_height=0.02,
                color=axes_color,
            )

            cad = []
            for cm in cad_mesh:
                cad.append(p.add_mesh(cm[0], color=cm[1], show_scalar_bar=False, opacity=cm[2]))

            if not show_geometry:
                p.add_checkbox_button_widget(toggle_vis_cad, value=False, position=(10, 70), size=30)
                toggle_vis_cad(False)
            else:
                p.add_checkbox_button_widget(toggle_vis_cad, value=True, position=(10, 70), size=30)

            p.add_text("Show Geometry", position=(70, 75), color=text_color, font_size=10)

        if output_file:
            p.show(auto_close=False, screenshot=output_file, full_screen=True)
        elif show:  # pragma: no cover
            p.show(auto_close=False, interactive=True)
        return p

    @pyaedt_function_handler()
    def __init_ffd(self, element_info):
        """Load far field information.

        Parameters
        ----------
        element_info : dict
            Information about the far fields imported.
            The keys of the dictionary represent the element names.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """

        for element, element_data in element_info.items():
            self.__raw_data[element] = {}
            self.__frequencies = []
            if os.path.exists(element_data["pattern_file"]):
                # Extract ports
                with open_file(element_data["pattern_file"], "r") as reader:
                    theta = [int(i) for i in reader.readline().split()]
                    phi = [int(i) for i in reader.readline().split()]
                reader.close()

                # Extract ffd information
                with open(element_data["pattern_file"], "r") as file:
                    ffd_text = file.read()

                segments = ffd_text.split("Frequency")
                eep_text_list = {}
                frequency_text_list = []

                for i, segment in enumerate(segments[1:]):
                    segment = segment.strip()
                    if segment:
                        lines = segment.split("\n")
                        frequency_text_list.append(lines[0])
                        eep_text_list[lines[0]] = lines[1:]

                theta_range = np.linspace(*theta)
                phi_range = np.linspace(*phi)

                for freq in frequency_text_list:
                    freq_hz = float(freq)
                    self.__frequencies.append(freq_hz)
                    temp_dict = {}
                    self.__raw_data[element][freq_hz] = {}
                    eep_txt = np.loadtxt(eep_text_list[freq])
                    Etheta = np.vectorize(complex)(eep_txt[:, 0], eep_txt[:, 1])
                    Ephi = np.vectorize(complex)(eep_txt[:, 2], eep_txt[:, 3])
                    temp_dict["Theta"] = theta_range
                    temp_dict["Phi"] = phi_range
                    temp_dict["rETheta"] = Etheta
                    temp_dict["rEPhi"] = Ephi
                    self.__raw_data[element][freq_hz] = temp_dict
            else:  # pragma: no cover
                raise Exception("Wrong far fields were imported.")

        return True

    @pyaedt_function_handler()
    def get_far_field_mesh(self, quantity="RealizedGain", quantity_format="dB10"):
        """Generate a PyVista ``UnstructuredGrid`` object that represents the far field mesh.

        Parameters
        ----------
        quantity : str, optional
            Far field quantity to plot. The default is ``"RealizedGain"``.
            Available quantities are: ``"RealizedGain"``, ``"RealizedGain_Phi"``, ``"RealizedGain_Theta"``,
            ``"rEPhi"``, ``"rETheta"``, and ``"rETotal"``.
        quantity_format : str, optional
            Conversion data function.
            Available functions are: ``"abs"``, ``"ang"``, ``"dB10"``, ``"dB20"``, ``"deg"``, ``"imag"``, ``"norm"``,
            and ``"real"``.

        Returns
        -------
        :class:`Pyvista.Plotter`
            ``UnstructuredGrid`` object representing the far field mesh.
        """
        farfield_data = self.farfield_data
        if quantity not in farfield_data:  # pragma: no cover
            raise Exception("Far field quantity is not available.")

        data = farfield_data[quantity]

        ff_data = conversion_function(data, quantity_format)

        if not isinstance(ff_data, np.ndarray):  # pragma: no cover
            raise Exception("Format of the quantity is wrong.")

        theta = np.deg2rad(np.array(farfield_data["Theta"]))
        phi = np.deg2rad(np.array(farfield_data["Phi"]))
        mesh = get_structured_mesh(theta=theta, phi=phi, ff_data=ff_data)
        return mesh

    @pyaedt_function_handler()
    def __get_geometry(self, off_screen=False):
        """Get 3D meshes."""
        model_info = self.metadata["model_info"]
        obj_meshes = []
        if self.__is_array:
            non_array_geometry = model_info.copy()
            components_info = self.__component_objects
            array_dimension = self.__array_dimension
            first_value = next(iter(model_info.values()))
            sf = AEDT_UNITS["Length"][first_value[3]]
            self.__model_units = first_value[3]
            cell_info = self.__cell_position

            for cell_row in cell_info:
                for cell_col in cell_row:
                    # Initialize an empty mesh for this component
                    model_pv = ModelPlotter()
                    model_pv.off_screen = off_screen
                    component_name = cell_col[0]
                    component_info = components_info[component_name]
                    rotation = cell_col[2]
                    for component_obj in component_info[1:]:
                        if component_obj in model_info:
                            if component_obj in non_array_geometry:
                                del non_array_geometry[component_obj]

                            cad_path = os.path.join(self.output_dir, model_info[component_obj][0])
                            if os.path.exists(cad_path):
                                model_pv.add_object(
                                    cad_path,
                                    model_info[component_obj][1],
                                    model_info[component_obj][2],
                                    model_info[component_obj][3],
                                )

                    model_pv.generate_geometry_mesh()
                    comp_meshes = []
                    row, col = cell_col[3]

                    # Perpendicular lattice vector
                    if self.__lattice_vector[0] != 0:
                        pos_x = (row - 1) * array_dimension[2] - array_dimension[0] / 2 + array_dimension[2] / 2
                        pos_y = (col - 1) * array_dimension[3] - array_dimension[1] / 2 + array_dimension[3] / 2
                    else:
                        pos_y = (row - 1) * array_dimension[2] - array_dimension[0] / 2 + array_dimension[2] / 2
                        pos_x = (col - 1) * array_dimension[3] - array_dimension[1] / 2 + array_dimension[3] / 2

                    for obj in model_pv.objects:
                        mesh = obj._cached_polydata
                        translated_mesh = mesh.copy()
                        color_cad = [i / 255 for i in obj.color]

                        translated_mesh.translate(
                            [-component_info[0][0] / sf, -component_info[0][1] / sf, -component_info[0][2] / sf],
                            inplace=True,
                        )

                        if rotation != 0:
                            translated_mesh.rotate_z(rotation, inplace=True)

                        # Translate the mesh to its position
                        translated_mesh.translate([pos_x / sf, pos_y / sf, component_info[0][2] / sf], inplace=True)

                        comp_meshes.append([translated_mesh, color_cad, obj.opacity])

                    obj_meshes.append(comp_meshes)
                    model_pv.close()

            obj_meshes = [item for sublist in obj_meshes for item in sublist]
        else:
            non_array_geometry = model_info

        if non_array_geometry:  # pragma: no cover
            model_pv = ModelPlotter()
            first_value = next(iter(non_array_geometry.values()))
            sf = AEDT_UNITS["Length"][first_value[3]]
            self.__model_units = first_value[3]
            model_pv.off_screen = off_screen
            for object_in in non_array_geometry.values():
                cad_path = os.path.join(self.output_dir, object_in[0])
                if os.path.exists(cad_path):
                    model_pv.add_object(
                        cad_path,
                        object_in[1],
                        object_in[2],
                        object_in[3],
                    )
                else:
                    self.logger.warning(f"{cad_path} does not exist.")
                    return False
            self.__model_units = first_value[3]
            model_pv.generate_geometry_mesh()
            i = 0
            for obj in model_pv.objects:
                mesh = obj._cached_polydata
                translated_mesh = mesh.copy()
                color_cad = [i / 255 for i in obj.color]

                if len(obj_meshes) > i:
                    obj_meshes[i][0] += translated_mesh
                else:
                    obj_meshes.append([translated_mesh, color_cad, obj.opacity])
                i += 1
            model_pv.close()

        return obj_meshes

    @pyaedt_function_handler()
    def get_port_index(self):
        """Get port indices.

        Returns
        -------
        list
            Element index.
        """
        port_index = {}

        port_name = self.all_element_names

        index_offset = 0
        if self.__is_array:
            port = port_name[0]
            first_index = port.split("[", 1)[1].split("]", 1)[0]
            if first_index[0] != "1":
                index_offset = int(float(first_index[0])) - 1

        for port in port_name:
            if self.__is_array:
                try:
                    str1 = port.split("[", 1)[1].split("]", 1)[0]
                    port_index[port] = [int(i) - index_offset for i in str1.split(",")]
                except Exception:
                    return False
            else:
                if not port_index:
                    port_index[port] = [1, 1]
                else:
                    last_value = list(port_index.values())[-1]
                    port_index[port] = [1, last_value[1] + 1]

        return port_index

    @staticmethod
    @pyaedt_function_handler()
    def __find_nearest(array, value):
        idx = np.searchsorted(array, value, side="left")
        if idx > 0 and (idx == len(array) or math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])):
            return idx - 1
        else:
            return idx

    @staticmethod
    @pyaedt_function_handler()
    def __rotation_to_euler_angles(rotation):  # pragma: no cover
        sy = math.sqrt(rotation[0, 0] * rotation[0, 0] + rotation[1, 0] * rotation[1, 0])
        singular = sy < 1e-6
        if not singular:
            x = math.atan2(rotation[2, 1], rotation[2, 2])
            y = math.atan2(-rotation[2, 0], sy)
            z = math.atan2(rotation[1, 0], rotation[0, 0])
        else:
            x = math.atan2(-rotation[1, 2], rotation[1, 1])
            y = math.atan2(-rotation[2, 0], sy)
            z = 0
        return np.array([x, y, z])


class FfdSolutionDataExporter:
    """Class to enable export of embedded element pattern data from HFSS.

    An instance of this class is returned from the
    :meth:`pyaedt.Hfss.get_antenna_data` method. This method allows creation of
    the embedded
    element pattern files for an antenna that have been solved in HFSS. The
    ``metadata_file`` properties can then be passed as arguments to
    instantiate an instance of the :class:`pyaedt.generic.farfield_visualization.FfdSolutionData` class for
    subsequent analysis and postprocessing of the array data.

    Note that this class is derived from the :class:`FfdSolutionData` class and can be used directly for
    far-field postprocessing and array analysis, but it remains a property of the
    :class:`pyaedt.Hfss` application.

    Parameters
    ----------
    app : :class:`pyaedt.Hfss`
        HFSS application instance.
    sphere_name : str
        Infinite sphere to use.
    setup_name : str
        Name of the setup. Make sure to build a setup string in the form of ``"SetupName : SetupSweep"``.
    frequencies : list
        Frequency list to export. Specify either a list of strings with units or a list of floats in Hertz units.
        For example, ``["9GHz", 9e9]``.
    variations : dict, optional
        Dictionary of all families including the primary sweep. The default value is ``None``.
    overwrite : bool, optional
        Whether to overwrite the existing far field solution data. The default is ``True``.
    export_touchstone : bool, optional
        Whether to export touchstone file. The default is ``False``. Working from 2024 R1.
    set_phase_center_per_port : bool, optional
        Set phase center per port location. The default is ``True``.

    Examples
    --------
    >>> import pyaedt
    >>> app = pyaedt.Hfss(version="2023.2", design="Antenna")
    >>> setup_name = "Setup1 : LastAdaptive"
    >>> frequencies = [77e9]
    >>> sphere = "3D"
    >>> data = app.get_antenna_data(frequencies,setup_name,sphere)
    >>> data.plot_3d(quantity_format="dB10")
    """

    def __init__(
        self,
        app,
        sphere_name,
        setup_name,
        frequencies,
        variations=None,
        overwrite=True,
        export_touchstone=True,
        set_phase_center_per_port=True,
    ):
        # Public
        self.sphere_name = sphere_name
        self.setup_name = setup_name

        if not variations:
            variations = app.available_variations.nominal_w_values_dict_w_dependent
        else:
            # Set variation to Nominal
            for var_name, var_value in variations.items():
                app[var_name] = var_value

        self.variations = variations
        self.overwrite = overwrite
        self.export_touchstone = export_touchstone
        if not isinstance(frequencies, list):
            self.frequencies = [frequencies]
        else:
            self.frequencies = frequencies

        # Private
        self.__app = app
        self.__model_info = {}
        self.__farfield_data = None
        self.__metadata_file = ""

        if self.__app.desktop_class.is_grpc_api and set_phase_center_per_port:
            self.__app.set_phase_center_per_port()
        else:  # pragma: no cover
            self.__app.logger.warning("Set phase center in port location manually.")

    @property
    def model_info(self):
        """List of models."""
        return self.__model_info

    @property
    def farfield_data(self):
        """Farfield data."""
        return self.__farfield_data

    @property
    def metadata_file(self):
        """Metadata file."""
        return self.__metadata_file

    @pyaedt_function_handler()
    def export_farfield(self):
        """Export far field solution data of each element."""

        # Output directory
        exported_name_map = "element.txt"
        solution_setup_name = self.setup_name.replace(":", "_").replace(" ", "")
        full_setup = "{}-{}".format(solution_setup_name, self.sphere_name)
        export_path = "{}/{}/".format(self.__app.working_directory, full_setup)
        local_path = "{}/{}/".format(settings.remote_rpc_session_temp_folder, full_setup)
        export_path = os.path.abspath(check_and_download_folder(local_path, export_path))

        # 2024.1
        file_path_xml = os.path.join(export_path, self.__app.design_name + ".xml")
        # 2023.2
        file_path_txt = os.path.join(export_path, exported_name_map)

        input_file = file_path_xml
        if self.__app.desktop_class.aedt_version_id < "2024.1":  # pragma: no cover
            input_file = file_path_txt

        # Create directory or check if files already exist
        if settings.remote_rpc_session:  # pragma: no cover
            settings.remote_rpc_session.filemanager.makedirs(export_path)
            file_exists = settings.remote_rpc_session.filemanager.pathexists(input_file)
        elif not os.path.exists(export_path):
            os.makedirs(export_path)
            file_exists = False
        else:
            file_exists = os.path.exists(input_file)

        time_before = time.time()

        # Export far field
        if self.overwrite or not file_exists:
            if self.__app.desktop_class.aedt_version_id < "2024.1":  # pragma: no cover
                is_exported = self.__app.export_element_pattern(
                    frequencies=self.frequencies,
                    setup=self.setup_name,
                    sphere=self.sphere_name,
                    variations=self.variations,
                    output_dir=export_path,
                )
                if not is_exported:  # pragma: no cover
                    return False
                if self.export_touchstone:
                    scattering = ScatteringMethods(self.__app)
                    setup_sweep_parts = self.setup_name.split(":")

                    setup_name = setup_sweep_parts[0].strip()
                    sweep_name = setup_sweep_parts[1].strip()

                    touchstone_file = scattering.export_touchstone(setup=setup_name, sweep=sweep_name)

                    if touchstone_file:
                        touchstone_name = os.path.basename(touchstone_file)
                        output_file = os.path.join(export_path, touchstone_name)
                        shutil.move(touchstone_file, output_file)
            else:
                is_exported = self.__app.export_antenna_metadata(
                    frequencies=self.frequencies,
                    setup=self.setup_name,
                    sphere=self.sphere_name,
                    variations=self.variations,
                    output_dir=export_path,
                    export_element_pattern=True,
                    export_objects=False,
                    export_touchstone=True,
                    export_power=True,
                )
                if not is_exported:  # pragma: no cover
                    return False
        else:
            self.__app.logger.info("Using existing element patterns files.")

        # Export geometry
        if os.path.isfile(input_file):
            geometry_path = os.path.join(export_path, "geometry")
            if not os.path.exists(geometry_path):
                os.mkdir(geometry_path)
            obj_list = self.__create_geometries(geometry_path)
            if obj_list:
                self.__model_info["object_list"] = obj_list

            if self.__app.component_array:
                component_array = self.__app.component_array[self.__app.component_array_names[0]]
                self.__model_info["component_objects"] = component_array.get_component_objects()
                self.__model_info["cell_position"] = component_array.get_cell_position()
                self.__model_info["array_dimension"] = [
                    component_array.a_length,
                    component_array.b_length,
                    component_array.a_length / component_array.a_size,
                    component_array.b_length / component_array.b_size,
                ]
                self.__model_info["lattice_vector"] = component_array.lattice_vector()

        # Create PyAEDT Metadata
        var = []
        if self.variations:
            for k, v in self.variations.items():
                var.append("{}='{}'".format(k, v))
            variation = " ".join(var)
        else:
            variation = self.__app.odesign.GetNominalVariation()

        power = {}

        if self.__app.desktop_class.aedt_version_id < "2024.1":

            available_categories = self.__app.post.available_quantities_categories()
            excitations = []
            is_power = True
            if "Active VSWR" in available_categories:  # pragma: no cover
                quantities = self.post.available_report_quantities(quantities_category="Active VSWR")
                for quantity in quantities:
                    excitations.append("ElementPatterns:=")
                    excitations.append(quantity.strip("ActiveVSWR(").strip(")"))
            elif "Terminal VSWR" in available_categories:
                quantities = self.__app.post.available_report_quantities(quantities_category="Terminal VSWR")
                for quantity in quantities:
                    excitations.append("ElementPatterns:=")
                    excitations.append(quantity.strip("VSWRt(").strip(")"))
                is_power = False
            elif "Gamma" in available_categories:
                quantities = self.__app.post.available_report_quantities(quantities_category="Gamma")
                for quantity in quantities:
                    excitations.append("ElementPatterns:=")
                    excitations.append(quantity.strip("Gamma(").strip(")"))
            else:  # pragma: no cover
                for excitation in self.__app.get_all_sources():
                    excitations.append("ElementPatterns:=")
                    excitations.append(excitation)
            for excitation_cont1 in range(len(excitations)):
                sources = {}
                incident_power = {}
                accepted_power = {}
                radiated_power = {}
                unit = "V"
                if is_power:
                    unit = "W"
                active_element = excitations[0]
                for excitation_cont2, port in enumerate(excitations):
                    if excitation_cont1 == excitation_cont2:
                        active_element = port
                        sources[port] = (f"1{unit}", "0deg")
                    else:
                        sources[port] = (f"0{unit}", "0deg")

                power[active_element] = {}

                self.__app.edit_sources(sources)

                report = self.__app.post.reports_by_category.antenna_parameters(
                    "IncidentPower", self.setup_name, self.sphere_name
                )
                data = report.get_solution_data()
                incident_powers = data.data_magnitude()

                report = self.__app.post.reports_by_category.antenna_parameters(
                    "RadiatedPower", self.setup_name, self.sphere_name
                )
                data = report.get_solution_data()
                radiated_powers = data.data_magnitude()

                report = self.__app.post.reports_by_category.antenna_parameters(
                    "AcceptedPower", self.setup_name, self.sphere_name
                )
                data = report.get_solution_data()
                accepted_powers = data.data_magnitude()

                for freq_cont, freq_str in enumerate(self.frequencies):
                    frequency = freq_str
                    if isinstance(freq_str, str):
                        frequency, units = decompose_variable_value(freq_str)
                        frequency = unit_converter(frequency, "Freq", units, "Hz")
                    incident_power[frequency] = incident_powers[freq_cont]
                    radiated_power[frequency] = radiated_powers[freq_cont]
                    accepted_power[frequency] = accepted_powers[freq_cont]

                power[active_element]["IncidentPower"] = incident_power
                power[active_element]["AcceptedPower"] = accepted_power
                power[active_element]["RadiatedPower"] = radiated_power

        pyaedt_metadata_file = FfdSolutionDataExporter.export_pyaedt_antenna_metadata(
            input_file=input_file, output_dir=export_path, variation=variation, model_info=self.model_info, power=power
        )
        if not pyaedt_metadata_file:  # pragma: no cover
            return False
        elapsed_time = time.time() - time_before
        self.__app.logger.info("Exporting embedded element patterns.... Done: %s seconds", elapsed_time)
        self.__metadata_file = pyaedt_metadata_file
        self.__farfield_data = FfdSolutionData(pyaedt_metadata_file)
        return pyaedt_metadata_file

    @staticmethod
    @pyaedt_function_handler()
    def export_pyaedt_antenna_metadata(
        input_file, output_dir, variation=None, model_info=None, power=None, touchstone_file=None
    ):
        """Obtain PyAEDT metadata JSON file from AEDT metadata XML file or embedded element pattern TXT file.

        Parameters
        ----------
        input_file : str
            Full path to the XML or TXT file.
        output_dir : str
            Full path to save the file to.
        variation : str, optional
            Label to identify corresponding variation.
        model_info : dict, optional
        power : dict, optional
            Dictionary with information of the incident power for each frequency.
            The default is ``None``, in which case an empty dictionary is applied.
            From AEDT 2024.1, this information is available from the XML input file.
            For example, the dictionary format for a two element farfield
            data = power[1000000000.0]["IncidentPower"]
            data = [1, 0.99]
        touchstone_file : str, optional
            Touchstone file name. The default is ``None``.

        Returns
        -------
        str
            Metadata JSON file.
        """
        from pyaedt.generic.touchstone_parser import find_touchstone_files

        if not variation:
            variation = "Nominal"

        if not power:
            power = {}

        if not touchstone_file:
            touchstone_file = ""

        pyaedt_metadata_file = os.path.join(output_dir, "pyaedt_antenna_metadata.json")
        items = {"variation": variation, "element_pattern": {}, "touchstone_file": touchstone_file}

        if os.path.isfile(input_file) and os.path.basename(input_file).split(".")[1] == "xml":
            # Metadata available from 2024.1
            antenna_metadata = FfdSolutionDataExporter.antenna_metadata(input_file)

            # Find all ffd files and move them to main directory
            for dir_path, _, filenames in os.walk(output_dir):
                ffd_files = [file for file in filenames if file.endswith(".ffd")]
                sNp_files = find_touchstone_files(dir_path)
                if ffd_files:
                    # Move ffd files to main directory
                    for ffd_file in ffd_files:
                        output_file = os.path.join(output_dir, ffd_file)
                        pattern_file = os.path.join(dir_path, ffd_file)
                        shutil.move(pattern_file, output_file)
                if sNp_files and not touchstone_file:
                    # Only one Touchstone allowed
                    sNp_name, sNp_path = next(iter(sNp_files.items()))
                    output_file = os.path.join(output_dir, sNp_name)
                    exported_touchstone_file = os.path.join(sNp_path)
                    shutil.move(exported_touchstone_file, output_file)
                    items["touchstone_file"] = sNp_name

            for metadata in antenna_metadata:

                incident_power = {}
                for i_freq, i_power_value in metadata["incident_power"].items():
                    frequency = i_freq
                    if isinstance(i_freq, str):
                        frequency, units = decompose_variable_value(i_freq)
                        if units:
                            frequency = unit_converter(frequency, "Freq", units, "Hz")
                    incident_power[frequency] = float(i_power_value)

                radiated_power = {}
                for i_freq, i_power_value in metadata["radiated_power"].items():
                    frequency = i_freq
                    if isinstance(i_freq, str):
                        frequency, units = decompose_variable_value(i_freq)
                        if units:
                            frequency = unit_converter(frequency, "Freq", units, "Hz")
                    radiated_power[frequency] = float(i_power_value)

                accepted_power = {}
                for i_freq, i_power_value in metadata["accepted_power"].items():
                    frequency = i_freq
                    if isinstance(i_freq, str):
                        frequency, units = decompose_variable_value(i_freq)
                        if units:
                            frequency = unit_converter(frequency, "Freq", units, "Hz")
                    accepted_power[frequency] = float(i_power_value)

                pattern = {
                    "file_name": metadata["file_name"],
                    "location": metadata["location"],
                    "incident_power": incident_power,
                    "radiated_power": radiated_power,
                    "accepted_power": accepted_power,
                }

                items["element_pattern"][metadata["name"]] = pattern
                pattern_file = os.path.join(output_dir, metadata["file_name"])
                if not os.path.isfile(pattern_file):  # pragma: no cover
                    return False

        elif os.path.isfile(input_file) and os.path.basename(input_file).split(".")[1] == "txt":

            # Find all ffd files and move them to main directory
            for dir_path, _, _ in os.walk(output_dir):
                sNp_files = find_touchstone_files(dir_path)
                if sNp_files and not touchstone_file:
                    # Only one Touchstone allowed
                    sNp_name, sNp_path = next(iter(sNp_files.items()))
                    output_file = os.path.join(output_dir, sNp_name)
                    exported_touchstone_file = os.path.join(sNp_path)
                    shutil.move(exported_touchstone_file, output_file)
                    items["touchstone_file"] = sNp_name
                    break

            with open_file(input_file, "r") as file:
                # Skip the first line
                file.readline()
                # Read and process the remaining lines
                for line in file:
                    antenna_metadata = line.strip().split()
                    if len(antenna_metadata) == 5:
                        element_name = antenna_metadata[0]
                        file_name = antenna_metadata[1]
                        if ".ffd" not in file_name:
                            file_name = file_name + ".ffd"
                        incident_power = None
                        radiated_power = None
                        accepted_power = None
                        if power:
                            incident_power = power[element_name]["IncidentPower"]
                            radiated_power = power[element_name]["RadiatedPower"]
                            accepted_power = power[element_name]["AcceptedPower"]

                        pattern = {
                            "file_name": file_name,
                            "location": [
                                float(antenna_metadata[2]),
                                float(antenna_metadata[3]),
                                float(antenna_metadata[4]),
                            ],
                            "incident_power": incident_power,
                            "radiated_power": radiated_power,
                            "accepted_power": accepted_power,
                        }
                        items["element_pattern"][antenna_metadata[0]] = pattern

        items["model_info"] = []
        if model_info:
            if "object_list" in model_info:
                items["model_info"] = model_info["object_list"]

            required_array_keys = ["array_dimension", "component_objects", "lattice_vector", "cell_position"]

            if all(key in model_info for key in required_array_keys):
                items["component_objects"] = model_info["component_objects"]
                items["cell_position"] = model_info["cell_position"]
                items["array_dimension"] = model_info["array_dimension"]
                items["lattice_vector"] = model_info["lattice_vector"]

        with open_file(pyaedt_metadata_file, "w") as f:
            json.dump(items, f, indent=2)
        return pyaedt_metadata_file

    @staticmethod
    @pyaedt_function_handler()
    def antenna_metadata(input_file):
        """Obtain metadata information from metadata XML file.

        Parameters
        ----------
        input_file : str
            Full path to the XML file.

        Returns
        -------
        dict
            Metadata information.

        """
        import xml.etree.ElementTree as ET  # nosec

        # Load the XML file
        tree = ET.parse(input_file)  # nosec
        root = tree.getroot()

        element_patterns = root.find("ElementPatterns")

        sources = []
        if element_patterns is None:  # pragma: no cover
            print("Element Patterns section not found in XML.")
        else:
            cont = 0
            # Iterate over each Source element
            for source in element_patterns.findall("Source"):
                source_info = {
                    "name": source.get("name"),
                    "file_name": source.find("Filename").text.strip(),
                    "location": source.find("ReferenceLocation").text.strip().split(","),
                }

                # Iterate over Power elements
                power_info = source.find("PowerInfo")
                if power_info is not None:
                    source_info["incident_power"] = {}
                    source_info["accepted_power"] = {}
                    source_info["radiated_power"] = {}
                    for power in power_info.findall("Power"):
                        freq = power.get("Freq")
                        source_info["incident_power"][freq] = {}
                        source_info["incident_power"][freq] = power.find("IncidentPower").text.strip()
                        source_info["accepted_power"][freq] = {}
                        source_info["accepted_power"][freq] = power.find("AcceptedPower").text.strip()
                        source_info["radiated_power"][freq] = {}
                        source_info["radiated_power"][freq] = power.find("RadiatedPower").text.strip()

                sources.append(source_info)
                cont += 1
        return sources

    @pyaedt_function_handler()
    def __create_geometries(self, export_path):
        """Export the geometry in OBJ format."""
        self.__app.logger.info("Exporting geometry...")
        model_pv = self.__app.post.get_model_plotter_geometries(plot_air_objects=False)
        obj_list = {}
        for obj in model_pv.objects:
            object_name = os.path.basename(obj.path)
            name = os.path.splitext(object_name)[0]
            original_path = os.path.dirname(obj.path)
            new_path = os.path.join(os.path.abspath(export_path), object_name)

            if not os.path.exists(new_path):
                new_path = shutil.move(obj.path, export_path)
            if os.path.exists(os.path.join(original_path, name + ".mtl")):  # pragma: no cover
                shutil.rmtree(os.path.join(original_path, name + ".mtl"), ignore_errors=True)
            obj_list[obj.name] = [
                os.path.join(os.path.basename(export_path), object_name),
                obj.color,
                obj.opacity,
                obj.units,
            ]
        return obj_list


class UpdateBeamForm:
    """Provides for updating far field data.

    This class is used to interact with the far field Theta and Phi scan.

    Parameters
    ----------
    farfield_data : :class:`pyaedt.generic.farfield_visualization.FfdSolutionData`
        Far field solution data instance.
    farfield_quantity : str, optional
        Quantity to plot. The default is ``"RealizedGain"``.
        Available quantities are: ``"RealizedGain"``, ``"RealizedGain_Phi"``, ``"RealizedGain_Theta"``,
        ``"rEPhi"``, ``"rETheta"``, and ``"rETotal"``.
    quantity_format : str, optional
        Conversion data function.
        Available functions are: ``"abs"``, ``"ang"``, ``"dB10"``, ``"dB20"``, ``"deg"``, ``"imag"``, ``"norm"``,
            and ``"real"``.
    """

    @pyaedt_function_handler(farfield_quantity="quantity")
    def __init__(self, ff, farfield_quantity="RealizedGain", quantity_format="abs"):
        self.output = ff._mesh
        self.__phi = 0
        self.__theta = 0
        self.ff = ff
        self.quantity = farfield_quantity
        self.quantity_format = quantity_format

    @pyaedt_function_handler()
    def __update_both(self):
        """Update far field."""
        self.ff.__farfield_data = self.ff.combine_farfield(phi_scan=self.__phi, theta_scan=self.__theta)
        self.ff._mesh = self.ff.get_far_field_mesh(self.quantity, self.quantity_format)
        self.output.copy_from(self.ff._mesh)

    @pyaedt_function_handler()
    def update_phi(self, phi):
        """Update the Phi value."""
        self.__phi = phi
        self.__update_both()

    @pyaedt_function_handler()
    def update_theta(self, theta):
        """Update the Theta value."""
        self.__theta = theta
        self.__update_both()
